import os
import pytest
import aristopy as ar

log = ar.Logger(default_log_level='WARNING').get_logger(__name__)

def test_export_and_plot():
    # Set a solver
    if hasattr(pytest, 'AVAILABLE_SOLVERS') and \
            len(pytest.AVAILABLE_SOLVERS) > 0:
        solver = pytest.AVAILABLE_SOLVERS[0]  # use first one from the list
    else:
        # if file is run as standalone test, pytest will not have the attribute
        # AVAILABLE_SOLVERS (it's set in test_000_...) --> try to use solver CBC
        solver = 'gurobi'
    log.info('The following solver is used in this test run: %s' % solver)

    es = ar.EnergySystem(number_of_time_steps=8, hours_per_time_step=1,
                         interest_rate=0, economic_lifetime=1)

    gas_src = ar.Source(ensys=es, name='gas_src', outlet=ar.Flow('F'),
                        commodity_cost=20)

    boiler = ar.Conversion(ensys=es, name='boiler', basic_variable='Q',
                           inlet=ar.Flow('F', 'gas_src'),
                           outlet=ar.Flow('Q', 'bus'),
                           capex_per_capacity=60e3, capacity_max=6,
                           user_expressions=['Q == F'])

    chp = ar.Conversion(ensys=es, name='chp', basic_variable='Q',
                        inlet=ar.Flow('F', 'gas_src'),
                        outlet=[ar.Flow('Q', 'bus'), ar.Flow('P', 'elec_snk')],
                        capex_per_capacity=600e3, has_operation_binary_var=True,
                        start_up_cost=10e3, min_load_rel=0.5,
                        capacity=4, user_expressions=['Q == 0.5 * F',
                                                      'P == 0.4 * F'])

    storage = ar.Storage(ensys=es, name='storage',
                         basic_variable='outlet_variable',
                         inlet=ar.Flow('Q', 'bus', 'Q_CHARGE'),
                         outlet=ar.Flow('Q', 'bus', 'Q_DISCHARGE'),
                         capex_per_capacity=1000, capacity_max=50,
                         opex_operation=1e-9, soc_initial=0.6,
                         self_discharge=0.05,
                         precise_inter_period_modeling=True)

    bus = ar.Bus(es, 'bus', inlet=ar.Flow('Q', var_name='Q_IN'),
                 outlet=ar.Flow('Q', var_name='Q_OUT'), losses=0)

    elec_snk = ar.Sink(ensys=es, name='elec_snk', inlet=ar.Flow('P'),
                       commodity_revenues=42)

    heat_snk = ar.Sink(ensys=es, name='heat_snk', inlet=ar.Flow('Q', 'bus'),
                       commodity_rate_fix=ar.Series('demand',
                                                    [4, 4, 6, 6, 10, 10, 4, 4]))

    # Helper function to write to or get from subdirectory 'temp'
    def temp_file(file_name):
        temp_dir = os.path.join(os.path.dirname(__file__), 'temp')
        if not os.path.exists(temp_dir):
            os.makedirs(temp_dir)
        return os.path.join(temp_dir, file_name)

    # Run the model as it is and assert the obj. fct. value
    es.optimize(declare_model=True, solver=solver, tee=False,
                results_file=temp_file('results.json'))

    assert es.model.Obj() == pytest.approx(-3.16985142475922e+06)

    plotter = ar.Plotter(json_file=temp_file('results.json'))

    plotter.plot_operation('bus', 'Q', level_of_detail=2, period_lw=0.5,
                           file_name=temp_file('bus'),
                           ylabel='Thermal energy transmission [MW/timestep]')

    plotter.plot_operation('storage', 'Q', level_of_detail=2, period_lw=0.5,
                           file_name=temp_file('storage'),
                           ylabel='Charging and discharging [MW] and SOC [MWh]')

    plotter.plot_operation('gas_src', 'F', level_of_detail=1, period_lw=0.5,
                           plot_single_period_with_index=0,
                           file_name=temp_file('fuel_single'))

    plotter.plot_objective(bar_width=0.8, bar_lw=0, show_plot=False,
                           file_name=temp_file('objective'))

    plotter.quick_plot('elec_snk', 'P', save_plot=True,
                       file_name=temp_file('quick_plot'))


if __name__ == '__main__':
    test_export_and_plot()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ==============================================================================
#    B U S
# ==============================================================================
"""
* File name: bus.py
* Last edited: 2020-06-14
* Created by: Stefan Bruche (TU Berlin)

A Bus component collects and transfers a commodity. Bus components can also
be used to model transmission lines between different sites.
"""
import pyomo.environ as pyo
from aristopy import utils
from aristopy.component import Component


class Bus(Component):
    def __init__(self, ensys, name, inlet, outlet,
                 basic_variable='inlet_variable',
                 has_existence_binary_var=False,
                 time_series_data=None, scalar_params=None,
                 additional_vars=None, user_expressions=None,
                 capacity=None, capacity_min=None, capacity_max=None,
                 capex_per_capacity=0, capex_if_exist=0,
                 opex_per_capacity=0, opex_if_exist=0, opex_operation=0,
                 losses=0
                 ):
        """
        Initialize an instance of the Bus class.

        .. note::
           See the documentation of the :class:`Component
           <aristopy.component.Component>` class for a description of all
           keyword arguments and inherited methods.

        :param losses: Factor to specify the relative loss of the transported
            commodity between inlet and outlet per hour of operation
            (0 => no loss; 1 => 100% loss).
            |br| *Default: 0*
        :type losses: float or int (0<=value<=1)
        """

        # Prevent None at inlet & outlet! (Flows are checked in Component init)
        if inlet is None:
            raise utils.io_error_message('Bus', name, 'inlet')
        if outlet is None:
            raise utils.io_error_message('Bus', name, 'outlet')

        Component.__init__(self, ensys=ensys, name=name,
                           inlet=inlet, outlet=outlet,
                           basic_variable=basic_variable,
                           has_existence_binary_var=has_existence_binary_var,
                           time_series_data=time_series_data,
                           scalar_params=scalar_params,
                           additional_vars=additional_vars,
                           user_expressions=user_expressions,
                           capacity=capacity, capacity_min=capacity_min,
                           capacity_max=capacity_max,
                           capex_per_capacity=capex_per_capacity,
                           capex_if_exist=capex_if_exist,
                           opex_per_capacity=opex_per_capacity,
                           opex_if_exist=opex_if_exist,
                           opex_operation=opex_operation
                           )

        # Check and set additional input arguments
        self.losses = utils.check_and_set_range_zero_one(losses, 'losses')

        # Store the names for the loading and unloading variables
        self.inlet_variable = self.inlet[0].var_name
        self.outlet_variable = self.outlet[0].var_name

        # Last step: Add the component to the EnergySystem instance
        self.add_to_energy_system(ensys, name)

    def __repr__(self):
        return '<Bus: "%s">' % self.name

    # ==========================================================================
    #    C O N V E N T I O N A L   C O N S T R A I N T   D E C L A R A T I O N
    # ==========================================================================
    def declare_component_constraints(self, ensys, model):
        """
        Method to declare all component constraints.

        The following constraint methods are inherited from the Component class
        and are not documented in this sub-class:

        * :meth:`con_couple_bi_ex_and_cap
          <aristopy.component.Component.con_couple_bi_ex_and_cap>`
        * :meth:`con_cap_min <aristopy.component.Component.con_cap_min>`

        *Method is not intended for public access!*

        :param ensys: Instance of the EnergySystem class
        :param model: Pyomo ConcreteModel of the EnergySystem instance
        """
        # Time-independent constraints :
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        self.con_couple_bi_ex_and_cap()
        self.con_cap_min()

        # Time-dependent constraints :
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        self.con_operation_limit(model)
        self.con_bus_balance(model)

    # **************************************************************************
    #    Time-dependent constraints
    # **************************************************************************
    def con_operation_limit(self, model):
        """
        The basic variable of a component is limited by its nominal capacity.
        This means, the operation (commodity at inlet or outlet) of a bus
        (MWh) is limited by its nominal power (MW) multiplied with the number of
        hours per time step. E.g.: |br|
        ``Q_IN[p, t] <= CAP * dt``

        *Method is not intended for public access!*
        """
        # Only required if component has a capacity variable
        if self.has_capacity_var:
            cap = self.variables[utils.CAP]['pyomo']
            basic_var = self.variables[self.basic_variable]['pyomo']
            dt = self.ensys.hours_per_time_step

            def con_operation_limit(m, p, t):
                return basic_var[p, t] <= cap * dt

            setattr(self.block, 'con_operation_limit', pyo.Constraint(
                model.time_set, rule=con_operation_limit))

    def con_bus_balance(self, model):
        """
        The quantity of the commodity at the outlet must equal the quantity at
        the inlet minus the the transmission loss share. A bus component cannot
        store a commodity (correction with "hours_per_time_step" not needed).
        E.g.: |br| ``Q_OUT[p, t] == Q_IN[p, t] * (1 - losses)``

        *Method is not intended for public access!*
        """
        inlet_var = self.variables[self.inlet_variable]['pyomo']
        outlet_var = self.variables[self.outlet_variable]['pyomo']

        def con_bus_balance(m, p, t):
            return outlet_var[p, t] == inlet_var[p, t] * (1 - self.losses)

        setattr(self.block, 'con_bus_balance', pyo.Constraint(
                model.time_set, rule=con_bus_balance))

    # ==========================================================================
    #    S E R I A L I Z E
    # ==========================================================================
    def serialize(self):
        """
        This method collects all relevant input data and optimization results
        from the Component instance, and returns them in an ordered dictionary.

        :return: OrderedDict
        """
        comp_dict = super().serialize()
        comp_dict['inlet_variable'] = self.inlet_variable
        comp_dict['outlet_variable'] = self.outlet_variable
        return comp_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ==============================================================================
#    C O N V E R S I O N
# ==============================================================================
"""
* File name: conversion.py
* Last edited: 2020-06-14
* Created by: Stefan Bruche (TU Berlin)

A conversion component takes commodities at the inlet and provides other
commodities at the outlet after an internal conversion.
"""
import pyomo.environ as pyo
from aristopy.component import Component
from aristopy import utils


class Conversion(Component):
    def __init__(self, ensys, name, inlet, outlet, basic_variable,
                 has_existence_binary_var=False, has_operation_binary_var=False,
                 time_series_data=None, scalar_params=None,
                 additional_vars=None, user_expressions=None,
                 capacity=None, capacity_min=None, capacity_max=None,
                 capacity_per_module=None, maximal_module_number=None,
                 capex_per_capacity=0, capex_if_exist=0,
                 opex_per_capacity=0, opex_if_exist=0, opex_operation=0,
                 start_up_cost=0, min_load_rel=None,
                 instances_in_group=1,
                 group_has_existence_order=True, group_has_operation_order=True,
                 use_inter_period_formulation=True
                 ):
        """
        Initialize an instance of the Conversion class.

        .. note::
           See the documentation of the :class:`Component
           <aristopy.component.Component>` class for a description of all
           keyword arguments and inherited methods.

        :param start_up_cost: Costs incurred when the state of the binary
            operation variable (BI_OP) changes from 0 (OFF) to 1 (ON) from one
            time step to the next one [€/Start]; (requires keyword argument
            'has_operation_binary_var' set to True).
            |br| *Default: 0*
        :type start_up_cost: float or int (>=0)

        :param min_load_rel: Value for the relative minimal part-load of a
            conversion unit (e.g., 0.5 represents 50% minimal load). Minimal
            part-loads require the availability of a binary operation variable
            ('has_operation_binary_var'=True) and currently, fixed capacities
            need to be specified ('capacity_min'='capacity_max').
            |br| *Default: None*
        :type min_load_rel: float or int (0<=value<=1), or None

        :param instances_in_group: States the number of similar component
            instances that are simultaneously created and arranged in a group.
            That means, the user has the possibility to instantiate multiple
            component instances (only for Conversion!) with identical
            specifications. These components work independently, but may have an
            order for their binary existence and/or operation variables (see:
            'group_has_existence_order', 'group_has_operation_order'). If a
            number larger than 1 is provided, the names of the components are
            extended with integers starting from 1 (e.g., 'conversion_1', ...).
            |br| *Default: 1*
        :type instances_in_group: int (>=0)

        :param group_has_existence_order: If multiple similar instances are
            created and arranged in a group ('instances_in_group'>1), the user
            might want to introduce an order of their binary existence variable
            values to break the symmetry of the optimization problem. If the
            flag is set to True a constraint is created for each component of
            the group that requires the previous component in the group to exist
            (BI_EX=1) before the own existence variable can take the value 1.
            |br| *Default: True*
        :type group_has_existence_order: bool

        :param group_has_operation_order: If multiple similar instances are
            created and arranged in a group ('instances_in_group'>1), the user
            might want to introduce an order of their binary operation variable
            values to break the symmetry of the optimization problem. If the
            flag is set to True a constraint is created for each component of
            the group and every time step of the optimization problem that
            requires the previous component in the group to be operated
            (BI_OP=1) before the own operation variable can take the value 1.
            |br| *Default: True*
        :type group_has_operation_order: bool

        :param use_inter_period_formulation: A second binary start-up variable
            is created (BI_SU_INTER), if the inter-period model formulation is
            requested (flag=True) and start-up cost are introduced (argument
            'start_up_cost'>0). This variable is used in case the model is
            optimized with aggregated time-series data (multiple periods). The
            variable links the binary operation variable values of otherwise
            independent periods, to enforce start-up cost if the operation
            status (BI_OP) changes from one period to the next (OFF to ON).
            Note: This formulation introduces additional variables and
            constraints, and increases both model accuracy and model complexity.
            |br| *Default: True*
        :type use_inter_period_formulation: bool
        """

        # Prevent None at inlet & outlet! (Flows are checked in Component init)
        if inlet is None:
            raise utils.io_error_message('Conversion', name, 'inlet')
        if outlet is None:
            raise utils.io_error_message('Conversion', name, 'outlet')

        Component.__init__(self, ensys=ensys, name=name,
                           inlet=inlet, outlet=outlet,
                           basic_variable=basic_variable,
                           has_existence_binary_var=has_existence_binary_var,
                           has_operation_binary_var=has_operation_binary_var,
                           time_series_data=time_series_data,
                           scalar_params=scalar_params,
                           additional_vars=additional_vars,
                           user_expressions=user_expressions,
                           capacity=capacity,
                           capacity_min=capacity_min,
                           capacity_max=capacity_max,
                           capacity_per_module=capacity_per_module,
                           maximal_module_number=maximal_module_number,
                           capex_per_capacity=capex_per_capacity,
                           capex_if_exist=capex_if_exist,
                           opex_per_capacity=opex_per_capacity,
                           opex_if_exist=opex_if_exist,
                           opex_operation=opex_operation
                           )

        # Check and set the value for the minimal relative part-load
        if min_load_rel is not None:
            utils.check_and_set_positive_number(min_load_rel, 'min_load_rel')
            assert min_load_rel <= 1, 'Maximal value for "min_load_rel" is 1!'
            if not self.has_bi_op:
                raise ValueError('Minimal part-loads require the availability '
                                 'of a binary operation variable '
                                 '("has_operation_binary_var=True").')
            if self.capacity is None:
                raise NotImplementedError(
                    'Minimal part-loads with flexible unit capacity can be '
                    'modelled but it requires some effort. This feature has'
                    ' not been implemented yet. Please consider the '
                    '"user_expressions" attribute to model it yourself.\n '
                    'E.g.: Q_CAP_OR_OFF[p, t] == CAP * BI_OP[p, t] * dt '
                    '--> Linearization required (Glover)!,\n '
                    'Q[p, t] >= Q_CAP_OR_OFF[p, t] * min_load_rel * dt')
        self.min_load_rel = min_load_rel

        # Check and set additional input arguments
        self.start_up_cost = utils.check_and_set_positive_number(
            start_up_cost, 'start_up_cost')
        self.use_inter_period_formulation = utils.check_and_set_bool(
            use_inter_period_formulation, 'use_inter_period_formulation')

        # Multiple instances formed and collected in one group
        utils.check_positive_int(instances_in_group, 'instances_in_group',
                                 zero_is_invalid=False)
        self.instances_in_group = instances_in_group
        self.group_has_existence_order = utils.check_and_set_bool(
            group_has_existence_order, 'group_has_existence_order')
        if self.instances_in_group > 1 and not self.has_bi_ex and \
                self.group_has_existence_order:  # is True
            raise ValueError('Group requires a binary existence variable if an '
                             'existence order is requested!')
        self.group_has_operation_order = utils.check_and_set_bool(
            group_has_operation_order, 'group_has_operation_order')
        if self.instances_in_group > 1 and not self.has_bi_op and \
                self.group_has_operation_order:  # is True
            raise ValueError('Group requires a binary operation variable if an '
                             'operation order is requested!')

        # Add variables for Start-up
        if self.start_up_cost != 0:
            if not self.has_bi_op:
                raise ValueError('Start-up cost require the availability '
                                 'of a binary operation variable.')
            # Else: add Start-up binary variable
            self.add_var(name=utils.BI_SU, domain='Binary', has_time_set=True)
            # If inter-period-formulation is requested: add another binary var
            if self.use_inter_period_formulation:
                self.add_var(name=utils.BI_SU_INTER, domain='Binary',
                             has_time_set=False, init=0,
                             alternative_set='inter_period_time_set')

        # Last step: Add the component to the EnergySystem instance
        self.add_to_energy_system(ensys, name, instances_in_group)

    def __repr__(self):
        return '<Conversion: "%s">' % self.name

    # ==========================================================================
    #    C O N V E N T I O N A L   C O N S T R A I N T   D E C L A R A T I O N
    # ==========================================================================
    def declare_component_constraints(self, ensys, model):
        """
        Method to declare all component constraints.

        The following constraint methods are inherited from the Component class
        and are not documented in this sub-class:

        * :meth:`con_couple_bi_ex_and_cap
          <aristopy.component.Component.con_couple_bi_ex_and_cap>`
        * :meth:`con_cap_min <aristopy.component.Component.con_cap_min>`
        * :meth:`con_cap_modular <aristopy.component.Component.con_cap_modular>`
        * :meth:`con_modular_sym_break
          <aristopy.component.Component.con_modular_sym_break>`
        * :meth:`con_couple_existence_and_modular
          <aristopy.component.Component.con_couple_existence_and_modular>`
        * :meth:`con_bi_var_ex_and_op_relation
          <aristopy.component.Component.con_bi_var_ex_and_op_relation>`
        * :meth:`con_couple_op_binary_and_basic_var
          <aristopy.component.Component.con_couple_op_binary_and_basic_var>`

        *Method is not intended for public access!*

        :param ensys: Instance of the EnergySystem class
        :param model: Pyomo ConcreteModel of the EnergySystem instance
        """
        # Time-independent constraints :
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        self.con_couple_bi_ex_and_cap()
        self.con_cap_min()
        self.con_cap_modular()
        self.con_modular_sym_break()
        self.con_couple_existence_and_modular()
        self.con_existence_sym_break(ensys)

        # Time-dependent constraints :
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        self.con_bi_var_ex_and_op_relation(model)
        self.con_couple_op_binary_and_basic_var(model)
        self.con_operation_limit(model)
        self.con_operation_sym_break(ensys, model)
        self.con_min_load_rel(model)
        self.con_start_up_cost(model)
        self.con_start_up_cost_inter(ensys, model)

    # **************************************************************************
    #    Time-independent constraints
    # **************************************************************************
    def con_existence_sym_break(self, ensys):
        """
        Constraint to state, that the next component in a group can only be
        built if the previous one already exists (symmetry break constraint for
        groups consisting of multiple components 'instances_in_group' > 1), if
        keyword argument 'group_has_existence_order' is set to True.
        E.g.: |br| ``BI_EX (of conversion_2) <= BI_EX (of conversion_1)``

        *Method is not intended for public access!*
        """
        if self.number_in_group > 1 and self.group_has_existence_order:
            bi_ex = self.variables[utils.BI_EX]['pyomo']
            # Get 'bi_ex' of the previous component in the sequence
            prior = self.group_name + '_{}'.format(self.number_in_group - 1)
            prior_comp = ensys.components[prior]
            bi_ex_prior = prior_comp.variables[utils.BI_EX]['pyomo']

            def con_existence_sym_break(m):
                return bi_ex <= bi_ex_prior
            setattr(self.block, 'con_existence_sym_break',
                    pyo.Constraint(rule=con_existence_sym_break))

    # **************************************************************************
    #    Time-dependent constraints
    # **************************************************************************
    def con_operation_sym_break(self, ensys, model):
        """
        Constraint to state, that the next component in a group can only be
        operated if the previous one is already 'ON' (symmetry break constraint
        for groups consisting of multiple components 'instances_in_group' > 1),
        if keyword argument 'group_has_operation_order' is set to True. E.g.:
        |br| ``BI_OP[p, t] (of conversion_2) <= BI_OP[p, t] (of conversion_1)``

        *Method is not intended for public access!*
        """
        if self.number_in_group > 1 and self.group_has_operation_order:
            bi_op = self.variables[utils.BI_OP]['pyomo']
            prior = self.group_name + '_{}'.format(self.number_in_group - 1)
            prior_comp = ensys.components[prior]
            bi_op_prior = prior_comp.variables[utils.BI_OP]['pyomo']

            def con_operation_sym_break(m, p, t):
                return bi_op[p, t] <= bi_op_prior[p, t]
            setattr(self.block, 'con_operation_sym_break', pyo.Constraint(
                model.time_set, rule=con_operation_sym_break))

    def con_operation_limit(self, model):
        """
        The basic variable of a component is limited by its nominal capacity.
        This usually means, the operation (main commodity) of a conversion
        (MWh) is limited by its nominal power (MW) multiplied with the number of
        hours per time step. E.g.: |br|
        ``Q[p, t] <= CAP * dt``

        *Method is not intended for public access!*
        """
        # Only required if component has a capacity variable
        if self.has_capacity_var:
            cap = self.variables[utils.CAP]['pyomo']
            basic_var = self.variables[self.basic_variable]['pyomo']
            has_time_set = self.variables[self.basic_variable]['has_time_set']
            dt = self.ensys.hours_per_time_step

            def con_operation_limit(m, p, t):
                if has_time_set:
                    return basic_var[p, t] <= cap * dt
                else:
                    # Exceptional case: Selection of a scalar basic variable
                    return basic_var == cap

            setattr(self.block, 'con_operation_limit', pyo.Constraint(
                model.time_set, rule=con_operation_limit))

    def con_min_load_rel(self, model):
        """
        Constraint to set a value for the relative minimal part-load of a
        conversion unit (e.g., a component can either be switched off (BI_OP=0)
        or can work between 50% and 100% of its nominal capacity if a value of
        0.5 for 'min_load_rel' is specified. The constraint requires the
        availability of a binary operation variable ('has_operation_binary_var'
        is True) and currently, fixed capacities need to be specified
        ('capacity_min'='capacity_max'='capacity'). E.g.: |br|
        ``Q[p, t] >= capacity * BI_OP[p, t] * min_load_rel * dt``

        *Method is not intended for public access!*
        """
        # Only required if minimal part-loads should be modelled.
        # Availability of binary operation variable and fixed capacity is
        # checked during component initialization.
        if self.min_load_rel is not None:
            bi_op = self.variables['BI_OP']['pyomo']
            basic_var = self.variables[self.basic_variable]['pyomo']
            cap = self.capacity
            min_load = self.min_load_rel
            dt = self.ensys.hours_per_time_step

            def con_min_load_rel(m, p, t):
                return basic_var[p, t] >= cap * bi_op[p, t] * min_load * dt
            setattr(self.block, 'con_min_load_rel', pyo.Constraint(
                model.time_set, rule=con_min_load_rel))

    def con_start_up_cost(self, model):
        """
        Constraint to determine the status of the binary start-up variable. If
        the operational status of a component changes from OFF (BI_OP=0) to ON
        (BI_OP=1) from one time step to the next, the binary start-up variable
        must take a value of 1. For shut-down and remaining ON / OFF, the status
        variable can take both values, but the obj. function forces it to be 0.
        E.g.: |br| ``0 <= BI_OP[t-1] - BI_OP[t] + BI_SU[t]``

        *Method is not intended for public access!*
        """
        # Only if the start-up cost value is larger than 0
        if self.start_up_cost > 0:
            bi_op = self.variables[utils.BI_OP]['pyomo']
            bi_su = self.variables[utils.BI_SU]['pyomo']

            def con_start_up_cost(m, p, t):
                if t != 0:  # not in first time step of a period
                    return 0 <= bi_op[p, t - 1] - bi_op[p, t] + bi_su[p, t]
                else:
                    return pyo.Constraint.Skip
            setattr(self.block, 'con_start_up_cost', pyo.Constraint(
                model.time_set, rule=con_start_up_cost))

    def con_start_up_cost_inter(self, ensys, model):
        """
        Constraint to link the binary operation variables of consecutive
        periods, to enforce start-up cost if the operation status (BI_OP)
        changes from one period to the next (OFF to ON).
        With this, the binary operation status of the last time step in the
        previous typical period and the operation status of the first time step
        in the current typical period are compared. Binary variable
        'BI_SU_INTER' indicates a potential status change from OFF to ON. E.g.:
        |br| ``0 <= BI_OP[p_typ-1, t_last] - BI_OP[p_typ, 0] + BI_SU_INTER[p]``

        *Method is not intended for public access!*
        """
        # Only if start-up cost are applied, the inter-period formulation is
        # requested and the data is clustered.
        if self.start_up_cost > 0 and self.use_inter_period_formulation \
                and ensys.is_data_clustered:
            bi_op = self.variables[utils.BI_OP]['pyomo']
            bi_su_inter = self.variables[utils.BI_SU_INTER]['pyomo']

            def con_start_up_cost_inter(m, p):
                # not for first period, because there is no precursor to use
                if not p == ensys.periods[0]:

                    typ_period = ensys.periods_order[p]
                    prev_typ_period = ensys.periods_order[p-1]
                    last_ts_idx = model.time_set.last()[1]

                    return 0 <= bi_op[prev_typ_period, last_ts_idx] - bi_op[
                        typ_period, 0] + bi_su_inter[p]
                else:
                    return pyo.Constraint.Skip
            setattr(self.block, 'con_start_up_cost_inter', pyo.Constraint(
                ensys.periods, rule=con_start_up_cost_inter))

    # ==========================================================================
    #    O B J E C T I V E   F U N C T I O N   C O N T R I B U T I O N
    # ==========================================================================
    def get_objective_function_contribution(self, ensys, model):
        """
        Calculate the objective function contributions of the component and add
        the values to the component dictionary "comp_obj_dict".

        *Method is not intended for public access!*

        :param ensys: Instance of the EnergySystem class
        :param model: Pyomo ConcreteModel of the EnergySystem instance
        """
        # Call method in "Component" class and calculate CAPEX and OPEX
        super().get_objective_function_contribution(ensys, model)

        # START-UP COST :
        # ~~~~~~~~~~~~~~~
        if self.start_up_cost > 0:
            bi_su = self.variables[utils.BI_SU]['pyomo']  # only avail. if '>0'
            start_cost_intra = -1 * ensys.pvf * self.start_up_cost * sum(
                bi_su[p, t] * ensys.period_occurrences[p] for p, t in
                model.time_set) / ensys.number_of_years
            if self.use_inter_period_formulation and ensys.is_data_clustered:
                bi_su_inter = self.variables[utils.BI_SU_INTER]['pyomo']
                start_cost_inter = -1 * ensys.pvf * self.start_up_cost * pyo.\
                    summation(bi_su_inter) / ensys.number_of_years
            else:
                start_cost_inter = 0
            self.comp_obj_dict['start_up_cost'] = \
                start_cost_intra + start_cost_inter

        return sum(self.comp_obj_dict.values())

    # ==========================================================================
    #    S E R I A L I Z E
    # ==========================================================================
    def serialize(self):
        """
        This method collects all relevant input data and optimization results
        from the Component instance, and returns them in an ordered dictionary.

        :return: OrderedDict
        """
        comp_dict = super().serialize()
        return comp_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ==============================================================================
#    S T O R A G E
# ==============================================================================
"""
* File name: storage.py
* Last edited: 2020-06-14
* Created by: Stefan Bruche (TU Berlin)

A storage component can collect a commodity at the inlet at one time step and
make it available at the outlet at another time step.
Thus, it is a component to provide flexibility.
"""
import pyomo.environ as pyo
from aristopy import utils
from aristopy.component import Component


class Storage(Component):
    def __init__(self, ensys, name, inlet, outlet,
                 basic_variable='inlet_variable',
                 has_existence_binary_var=False,
                 time_series_data=None, scalar_params=None,
                 additional_vars=None, user_expressions=None,
                 capacity=None, capacity_min=None, capacity_max=None,
                 capacity_per_module=None, maximal_module_number=None,
                 capex_per_capacity=0, capex_if_exist=0,
                 opex_per_capacity=0, opex_if_exist=0, opex_operation=0,
                 charge_rate=1, discharge_rate=1, self_discharge=0,
                 charge_efficiency=1, discharge_efficiency=1,
                 soc_min=0, soc_max=1, soc_initial=None,
                 use_inter_period_formulation=True,
                 precise_inter_period_modeling=False
                 ):
        """
        Initialize an instance of the Storage class.

        .. note::
           See the documentation of the :class:`Component
           <aristopy.component.Component>` class for a description of all
           keyword arguments and inherited methods.

        :param charge_rate: Ratio between the maximum charging power or flow and
            the storage capacity. It indicates the reciprocal value of the
            time for a full storage charging process from empty to full (e.g.,
            'charge_rate'=1/6 => 6 hours needed to load an empty storage fully).
            |br| *Default: 1*
        :type charge_rate: float or int (>=0)

        :param discharge_rate: Ratio between the maximum discharging power or
            flow and the storage capacity. It indicates the reciprocal value of
            the time for a full storage discharging process from full to empty
            (e.g., 'discharge_rate'=1/6 => 6 hours needed for emptying a fully
            loaded storage).
            |br| *Default: 1*
        :type discharge_rate: float or int (>=0)

        :param self_discharge: Share of the storage content that is dissipated
            and can not be used (e.g., heat losses to the environment for a
            heat storage). The value is specified in "percent per hour" [%/h].
            |br| *Default: 0*
        :type self_discharge: float or int (0<=value<=1)

        :param charge_efficiency: Efficiency value for the charging process.
            It indicates the ratio between stored and entering commodities.
            E.g., 'charge_efficiency'=0.9 => for 1 MWh entering the storage in
            one time step 0.9 MWh are stored, and 0.1 MWh are lost.
            |br| *Default: 1*
        :type charge_efficiency: float or int (0<=value<=1)

        :param discharge_efficiency: Efficiency of the discharging process.
            It indicates the ratio between the usable commodity at the outlet
            and the reduction in the stored commodity. E.g.,
            'discharge_efficiency'=0.9 => if SOC is reduced by 1 MWh in one time
            step, 0.9 MWh are available at the outlet, and 0.1 MWh are lost.
            |br| *Default: 1*
        :type discharge_efficiency: float or int (0<=value<=1)

        :param soc_min: Relative value to provide a lower bound for the usable
            storage capacity. E.g., with a storage capacity of 5 MWh and a value
            for 'soc_min' given with 0.2, the SOC cannot fall below 1 MWh.
            |br| *Default: 0*
        :type soc_min: float or int (0<=value<=1)

        :param soc_max: Relative value to provide an upper bound for the usable
            storage capacity. E.g., with a storage capacity of 5 MWh and a value
            for 'soc_max' given with 0.8, the SOC cannot exceed 4 MWh.
            |br| *Default: 1*
        :type soc_max: float or int (0<=value<=1)

        :param soc_initial: Provides a value for the relative state of charge in
            the first time step of the optimization problem (e.g., 0.5 => 50%).
            The initial SOC value is applied to all periods if the model has
            multiple periods (calculation with clustered data), and the keyword
            argument 'use_inter_period_formulation' is set to False.
            |br| *Default: None*
        :type soc_initial: float or int (0<=value<=1), or None

        :param use_inter_period_formulation: States whether a model formulation
            should be applied that connects the states of charge of a storage
            component for otherwise independent periods (only used if time
            series aggregation is applied). Additional variables and constraints
            are created if the keyword argument is set to True. This formulation
            enables the (energy) transport between periods (especially relevant
            for long-term storages) and likewise increases the model complexity.
            |br| *Default: True*
        :type use_inter_period_formulation: bool

        :param precise_inter_period_modeling: States whether the inter-period
            formulation should be implemented in a simplified (False) or precise
            (True) way. The type of formulation influences how the constraints
            are modeled that enforce the SOC's bounds. If the storage has only a
            low self-discharge value, it is recommended to choose the simplified
            version (False). This version introduces some additional variables,
            but requires a significantly smaller number of constraints
            (`Ref: DOI 10.1016/j.apenergy.2018.01.023
            <https://doi.org/10.1016/j.apenergy.2018.01.023>`_).
            |br| *Default: False*
        :type precise_inter_period_modeling: bool
        """

        # Prevent None at inlet & outlet! (Flows are checked in Component init)
        if inlet is None:
            raise utils.io_error_message('Storage', name, 'inlet')
        if outlet is None:
            raise utils.io_error_message('Storage', name, 'outlet')

        # Set an upper bound for the storage capacity if nothing is specified to
        # make sure that storage components always have a capacity variable!
        if not capacity and not capacity_min and not capacity_max:
            capacity_max = 1e9

        Component.__init__(self, ensys=ensys, name=name,
                           inlet=inlet, outlet=outlet,
                           basic_variable=basic_variable,
                           has_existence_binary_var=has_existence_binary_var,
                           time_series_data=time_series_data,
                           scalar_params=scalar_params,
                           additional_vars=additional_vars,
                           user_expressions=user_expressions,
                           capacity=capacity,
                           capacity_min=capacity_min,
                           capacity_max=capacity_max,
                           capacity_per_module=capacity_per_module,
                           maximal_module_number=maximal_module_number,
                           capex_per_capacity=capex_per_capacity,
                           capex_if_exist=capex_if_exist,
                           opex_per_capacity=opex_per_capacity,
                           opex_if_exist=opex_if_exist,
                           opex_operation=opex_operation
                           )

        # Check and set additional input arguments
        self.charge_rate = utils.check_and_set_positive_number(
            charge_rate, 'charge_rate')
        self.discharge_rate = utils.check_and_set_positive_number(
            discharge_rate, 'discharge_rate')
        self.self_discharge = utils.check_and_set_range_zero_one(
            self_discharge, 'self_discharge')
        self.charge_efficiency = utils.check_and_set_range_zero_one(
            charge_efficiency, 'charge_efficiency')
        self.discharge_efficiency = utils.check_and_set_range_zero_one(
            discharge_efficiency, 'discharge_efficiency')
        self.soc_min = utils.check_and_set_range_zero_one(soc_min, 'soc_min')
        self.soc_max = utils.check_and_set_range_zero_one(soc_max, 'soc_max')
        self.soc_initial = utils.check_and_set_range_zero_one(
            soc_initial, 'soc_initial') if soc_initial is not None else None

        self.use_inter_period_formulation = utils.check_and_set_bool(
            use_inter_period_formulation, 'use_inter_period_formulation')
        self.precise_inter_period_modeling = utils.check_and_set_bool(
            precise_inter_period_modeling, 'precise_inter_period_modeling')

        # Store the names of the charging and discharging variables
        self.charge_variable = self.inlet[0].var_name
        self.discharge_variable = self.outlet[0].var_name

        # Create a state of charge (SOC) variable and if the inter-period
        # formulation is selected create an additional inter-period SOC variable
        # and SOC_MIN and SOC_MAX variables if precise modeling is requested.
        if not self.use_inter_period_formulation:
            self.add_var(utils.SOC, has_time_set=False,
                         alternative_set='intra_period_time_set')  # NonNegReal
        # use inter-period formulation:
        else:
            self.add_var(utils.SOC, domain='Reals', has_time_set=False,
                         alternative_set='intra_period_time_set')  # Real
            self.add_var(utils.SOC_INTER, has_time_set=False,
                         alternative_set='inter_period_time_set')
            if not self.precise_inter_period_modeling:
                self.add_var(utils.SOC_MAX, has_time_set=False,
                             alternative_set='typical_periods_set',
                             domain='Reals')
                self.add_var(utils.SOC_MIN, has_time_set=False,
                             alternative_set='typical_periods_set',
                             domain='Reals')

        # Last step: Add the component to the EnergySystem instance
        self.add_to_energy_system(ensys, name)

    def __repr__(self):
        return '<Storage: "%s">' % self.name

    # ==========================================================================
    #    C O N V E N T I O N A L   C O N S T R A I N T   D E C L A R A T I O N
    # ==========================================================================
    def declare_component_constraints(self, ensys, model):
        """
        Method to declare all component constraints.

        The following constraint methods are inherited from the Component class
        and are not documented in this sub-class:

        * :meth:`con_couple_bi_ex_and_cap
          <aristopy.component.Component.con_couple_bi_ex_and_cap>`
        * :meth:`con_cap_min <aristopy.component.Component.con_cap_min>`
        * :meth:`con_cap_modular <aristopy.component.Component.con_cap_modular>`
        * :meth:`con_modular_sym_break
          <aristopy.component.Component.con_modular_sym_break>`
        * :meth:`con_couple_existence_and_modular
          <aristopy.component.Component.con_couple_existence_and_modular>`

        *Method is not intended for public access!*

        :param ensys: Instance of the EnergySystem class
        :param model: Pyomo ConcreteModel of the EnergySystem instance
        """
        # Time-independent constraints :
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        self.con_couple_bi_ex_and_cap()
        self.con_cap_min()
        self.con_cap_modular()
        self.con_modular_sym_break()
        self.con_couple_existence_and_modular()

        # Time-dependent constraints :
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        self.con_operation_limit(model)
        self.con_soc_balance(ensys, model)
        self.con_charge_rate(ensys, model)
        self.con_discharge_rate(ensys, model)
        self.con_cyclic_condition(ensys, model)
        self.con_soc_initial(ensys, model)
        self.con_soc_intra_period_start(ensys, model)  # only if inter-period f.
        self.con_soc_inter_period_balance(ensys, model)  # dito
        self.con_soc_bounds_without_inter_period_formulation(ensys, model)
        self.con_soc_bounds_with_inter_period_formulation_simple(ensys, model)
        self.con_soc_bounds_with_inter_period_formulation_precise(ensys, model)

    # **************************************************************************
    #    Time-dependent constraints
    # **************************************************************************
    def con_operation_limit(self, model):
        """
        The state of charge (SOC) of a storage component is limit by its nominal
        capacity in all time steps. E.g.: |br|
        ``SOC[p, t] <= CAP``

        *Method is not intended for public access!*
        """
        cap = self.variables[utils.CAP]['pyomo']
        soc = self.variables[utils.SOC]['pyomo']

        def con_operation_limit(m, p, t):
            return soc[p, t] <= cap
        setattr(self.block, 'con_operation_limit', pyo.Constraint(
            model.time_set, rule=con_operation_limit))

    def con_soc_balance(self, ensys, model):
        """
        Constraint that connects the state of charge (SOC) of each time step
        with the charging and discharging events. The change in the SOC between
        two points in time has to match the values of charging and discharging
        and the self-discharge of the storage (explicit Euler formulation).
        Note: The SOC is not necessarily a value between 0 and 1 here. E.g.,
        |br| ``SOC[p, t+1] <= SOC[p, t]*(1-self_dischar)**dt
        + Q_IN[p, t]*eta_char - Q_OUT[p, t]/eta_dischar``

        *Method is not intended for public access!*
        """
        charge = self.variables[self.charge_variable]['pyomo']
        discharge = self.variables[self.discharge_variable]['pyomo']
        soc = self.variables[utils.SOC]['pyomo']
        dt = ensys.hours_per_time_step

        def con_soc_balance(m, p, t):
            return soc[p, t+1] == soc[p, t] * (1-self.self_discharge)**dt \
                   + charge[p, t] * self.charge_efficiency \
                   - discharge[p, t] / self.discharge_efficiency
        setattr(self.block, 'con_soc_balance', pyo.Constraint(
            model.time_set, rule=con_soc_balance))

    def con_charge_rate(self, ensys, model):
        """
        Constraint to limit the value of the charge variable by applying a
        maximal charge rate. E.g.: |br|
        ``Q_IN[p, t] <= CAP * charge_rate * dt``

        *Method is not intended for public access!*
        """
        cap = self.variables[utils.CAP]['pyomo']
        charge = self.variables[self.charge_variable]['pyomo']
        dt = ensys.hours_per_time_step

        def con_charge_rate(m, p, t):
            return charge[p, t] <= cap * dt * self.charge_rate
        setattr(self.block, 'con_charge_rate', pyo.Constraint(
            model.time_set, rule=con_charge_rate))

    def con_discharge_rate(self, ensys, model):
        """
        Constraint to limit the value of the discharge variable by applying a
        maximal discharge rate. E.g.: |br|
        ``Q_OUT[p, t] <= CAP * discharge_rate * dt``

        *Method is not intended for public access!*
        """
        cap = self.variables[utils.CAP]['pyomo']
        discharge = self.variables[self.discharge_variable]['pyomo']
        dt = ensys.hours_per_time_step

        def con_discharge_rate(m, p, t):
            return discharge[p, t] <= cap * dt * self.discharge_rate
        setattr(self.block, 'con_discharge_rate', pyo.Constraint(
            model.time_set, rule=con_discharge_rate))

    def con_cyclic_condition(self, ensys, model):
        """
        Constraint to enforce that the SOC in the last time step of a period
        (after charging and discharging events) equals the SOC at the beginning
        of the same period.
        In case the inter-period formulation is activated, this constraint
        demands that the cycle condition is also fulfilled for the full time
        scale problem => SOC in global first time step (e.g., SOC[t=1]) equals
        SOC in global last time step (e.g., SOC[t=8760]).

        *Method is not intended for public access!*
        """
        if self.use_inter_period_formulation and ensys.is_data_clustered:
            soc_inter = self.variables[utils.SOC_INTER]['pyomo']

            def con_cyclic_condition_inter(m):
                last_idx = model.inter_period_time_set.last()
                return soc_inter[0] == soc_inter[last_idx]
            setattr(self.block, 'con_cyclic_condition_inter', pyo.Constraint(
                rule=con_cyclic_condition_inter))
        else:
            # Use the formulation without inter-period time steps. This version
            # is computationally less challenging. All periods represent
            # independent entities. Energy cannot be transferred between periods
            # Only one "typical period" exists if data is not clustered [0]
            soc = self.variables[utils.SOC]['pyomo']

            def con_cyclic_condition(m, p):
                last_t_idx = model.intra_period_time_set.last()[1]
                return soc[p, 0] == soc[p, last_t_idx]
            setattr(self.block, 'con_cyclic_condition', pyo.Constraint(
                model.typical_periods_set, rule=con_cyclic_condition))

    def con_soc_initial(self, ensys, model):
        """
        Constraint that sets a value for the relative state of charge in the
        first time step of the optimization problem. The same initial SOC value
        is applied to all periods if the model has multiple periods. Otherwise,
        the initial value is only specified for very beginning of the time set.
        E.g.: |br|
        ``SOC[p, 0] == CAP * soc_initial`` or |br|
        ``SOC_INTER[0] == CAP * soc_initial``

        *Method is not intended for public access!*
        """
        if self.soc_initial is not None:
            cap = self.variables[utils.CAP]['pyomo']

            if self.use_inter_period_formulation and ensys.is_data_clustered:
                soc_inter = self.variables[utils.SOC_INTER]['pyomo']

                def con_soc_inter_initial(m):
                    return soc_inter[0] == cap * self.soc_initial
                setattr(self.block, 'con_soc_inter_initial', pyo.Constraint(
                    rule=con_soc_inter_initial))
            else:
                # Use the formulation without inter-period time steps. This
                # is computationally less challenging. All periods represent
                # independent entities. Energy can't be transferred between them
                # Only one "typical period" exists if data is not clustered [0]
                soc = self.variables[utils.SOC]['pyomo']

                def con_soc_initial(m, p):
                    return soc[p, 0] == cap * self.soc_initial
                setattr(self.block, 'con_soc_initial', pyo.Constraint(
                    model.typical_periods_set, rule=con_soc_initial))

    def con_soc_intra_period_start(self, ensys, model):
        """
        The state of charge consists of two parts (Intra and Inter), if the
        intra-period model formulation ('use_inter_period_formulation'=True) is
        selected. This constraint sets the intra-period part (SOC) to zero at
        the beginning of each period. E.g.: |br|
        ``SOC[p, 0] == 0``

        *Method is not intended for public access!*
        """
        if self.use_inter_period_formulation and ensys.is_data_clustered:
            soc = self.variables[utils.SOC]['pyomo']

            def con_soc_intra_period_start(m, p):
                return soc[p, 0] == 0
            setattr(self.block, 'con_soc_intra_period_start', pyo.Constraint(
                model.typical_periods_set, rule=con_soc_intra_period_start))

    def con_soc_inter_period_balance(self, ensys, model):
        """
        Constraint to calculate the inter-period state of charge (SOC_INTER) of
        the next period from the previous one and the value of the SOC (Intra)
        in the last time step of the related typical period. The constraint
        also accounts for self-discharge losses associated with the variable
        SOC_INTER (=> this can result in steps in the overall SOC profile at the
        boundary of 2 periods). E.g.: |br|
        ``SOC_INTER[p+1] == SOC_INTER[p]*(1-self_dischar)**(
        time_steps_per_period*dt) + SOC[typ_p, last_t]``

        *Method is not intended for public access!*
        """
        if self.use_inter_period_formulation and ensys.is_data_clustered:
            soc = self.variables[utils.SOC]['pyomo']
            soc_inter = self.variables[utils.SOC_INTER]['pyomo']
            dt = ensys.hours_per_time_step

            def con_soc_inter_period_balance(m, p):
                typ_period = ensys.periods_order[p]
                last_ts_idx = model.intra_period_time_set.last()[1]
                return soc_inter[p + 1] == soc_inter[p] * (
                        1 - self.self_discharge)**(
                        ensys.number_of_time_steps_per_period * dt) + soc[
                    typ_period, last_ts_idx]

            setattr(self.block, 'con_soc_inter_period_balance',
                    pyo.Constraint(ensys.periods,
                                   rule=con_soc_inter_period_balance))

    def con_soc_bounds_without_inter_period_formulation(self, ensys, model):
        """
        Two constraints that specify bounds for the state of charge variable
        (minimal and maximal values) in case the inter-period formulation is NOT
        selected, or the data is NOT clustered!  E.g.: |br|
        ``SOC[p, t] >= CAP * soc_min`` and |br|
        ``SOC[p, t] <= CAP * soc_max``

        *Method is not intended for public access!*
        """
        if not self.use_inter_period_formulation or not ensys.is_data_clustered:
            soc = self.variables[utils.SOC]['pyomo']
            cap = self.variables[utils.CAP]['pyomo']

            # Always enforce a minimal value for the SOC (CAP*soc_min) with a
            # constraint, because it is possible that the domain of the SOC
            # variable is 'Reals' (if the storage component is constructed with
            # 'use_inter_period_formulation'=True, but the time series
            # aggregation was not performed before optimization).
            def con_minimal_soc(m, p, t):
                return soc[p, t] >= cap * self.soc_min
            setattr(self.block, 'con_minimal_soc', pyo.Constraint(
                model.intra_period_time_set, rule=con_minimal_soc))

            # Only built the constraint if the value for soc_max is less than 1.
            # Otherwise it is already considered by "con_operation_limit".
            if self.soc_max < 1:
                def con_maximal_soc(m, p, t):
                    return soc[p, t] <= cap * self.soc_max
                setattr(self.block, 'con_maximal_soc', pyo.Constraint(
                    model.intra_period_time_set, rule=con_maximal_soc))

    def con_soc_bounds_with_inter_period_formulation_simple(self, ensys, model):
        """
        Four constraints that define the bounds for the state of charge variable
        (minimal and maximal values) in a simplified way in case the inter-
        period formulation is requested and the data is clustered. The error is
        relatively small in comparison to the precise method if the specified
        value for the 'self_discharge' is not too high. However, this version
        requires a reasonable smaller number of constraints but also some
        additional variables. |br| See Eq. B1 and B2 in the Appendix of
        `Ref: DOI 10.1016/j.apenergy.2018.01.023
        <https://doi.org/10.1016/j.apenergy.2018.01.023>`_.

        *Method is not intended for public access!*
        """
        if self.use_inter_period_formulation and ensys.is_data_clustered \
                and not self.precise_inter_period_modeling:

            # Get variables:
            soc = self.variables[utils.SOC]['pyomo']
            cap = self.variables[utils.CAP]['pyomo']
            soc_max = self.variables[utils.SOC_MAX]['pyomo']
            soc_min = self.variables[utils.SOC_MIN]['pyomo']
            soc_inter = self.variables[utils.SOC_INTER]['pyomo']
            dt = ensys.hours_per_time_step

            # The variable "SOC_MAX" of a typical period is larger than all
            # occurring intra-period SOCs in the same period (except the last
            # one since it is already used in "SOC_INTER" of the next period)
            def con_soc_max_intra(m, p, t):
                return soc[p, t] <= soc_max[p]
            setattr(self.block, 'con_soc_max_intra', pyo.Constraint(
                model.time_set, rule=con_soc_max_intra))

            # The variable "SOC_MIN" of a typical period is smaller than all
            # occurring intra-period SOCs in the same period (except the last
            # one since it is already used in "SOC_INTER" of the next period)
            def con_soc_min_intra(m, p, t):
                return soc[p, t] >= soc_min[p]
            setattr(self.block, 'con_soc_min_intra', pyo.Constraint(
                model.time_set, rule=con_soc_min_intra))

            # The inter-period SOC at the beginning of a period plus the maximum
            # intra-period SOC of the associated typical period is less than
            # the available capacity times the maximal usable share (soc_max).
            def con_soc_max_bound_simple(m, p):
                typ_period = ensys.periods_order[p]
                return soc_inter[p] + soc_max[typ_period] <= cap * self.soc_max
            setattr(model, 'con_soc_max_bound_simple', pyo.Constraint(
                ensys.periods, rule=con_soc_max_bound_simple))

            # The inter-period SOC at the beginning of a period minus the
            # maximal self discharge during that period plus the minimum
            # intra-period SOC of the associated typical period is larger than
            # the available capacity times the minimal required SOC (soc_min).
            def con_soc_min_bound_simple(m, p):
                typ_period = ensys.periods_order[p]
                return soc_inter[p] * (1 - self.self_discharge)**(
                        ensys.number_of_time_steps_per_period * dt) + soc_min[
                    typ_period] >= cap * self.soc_min
            setattr(model, 'con_soc_min_bound_simple', pyo.Constraint(
                ensys.periods, rule=con_soc_min_bound_simple))

    def con_soc_bounds_with_inter_period_formulation_precise(self, ensys,
                                                             model):
        """
        Two constraints define the bounds for the state of charge variable
        (minimal and maximal values) in a precise way if the inter-period
        formulation is requested, and the data is clustered.
        This version requires two constraints for each time step of the full-
        scale problem and can be computationally expensive. Users might want to
        consider applying the simplified constraint formulation (keyword
        argument 'precise_inter_period_modeling'=False).
        |br| See Eq. 20 in `Ref: DOI 10.1016/j.apenergy.2018.01.023
        <https://doi.org/10.1016/j.apenergy.2018.01.023>`_.

        *Method is not intended for public access!*
        """
        if self.use_inter_period_formulation and ensys.is_data_clustered \
                and self.precise_inter_period_modeling:

            # Get variables:
            soc = self.variables[utils.SOC]['pyomo']
            cap = self.variables[utils.CAP]['pyomo']
            soc_inter = self.variables[utils.SOC_INTER]['pyomo']
            dt = ensys.hours_per_time_step

            # The inter-period SOC at the beginning of a period minus the self
            # discharge plus the intra-period SOC in each hour of that period
            # is less than the available capacity times the max. usable share.
            def con_soc_max_bound_precise(m, p, t):
                typ_period = ensys.periods_order[p]
                return soc_inter[p] * (1 - self.self_discharge)**(t * dt) + soc[
                    typ_period, t] <= cap * self.soc_max
            setattr(model, 'con_soc_max_bound_precise', pyo.Constraint(
                ensys.periods, range(ensys.number_of_time_steps_per_period),
                rule=con_soc_max_bound_precise))

            # The inter-period SOC at the beginning of a period minus the self
            # discharge plus the intra-period SOC in each hour of that period
            # is larger than the available capacity times the min. required SOC.
            def con_soc_min_bound_precise(m, p, t):
                typ_period = ensys.periods_order[p]
                return soc_inter[p] * (1 - self.self_discharge)**(t * dt) + soc[
                    typ_period, t] >= cap * self.soc_min
            setattr(model, 'con_soc_min_bound_precise', pyo.Constraint(
                ensys.periods, range(ensys.number_of_time_steps_per_period),
                rule=con_soc_min_bound_precise))

    # ==========================================================================
    #    S E R I A L I Z E
    # ==========================================================================
    def serialize(self):
        """
        This method collects all relevant input data and optimization results
        from the Component instance, and returns them in an ordered dictionary.

        :return: OrderedDict
        """
        comp_dict = super().serialize()
        comp_dict['charge_variable'] = self.charge_variable
        comp_dict['discharge_variable'] = self.discharge_variable
        comp_dict['soc_variable'] = utils.SOC
        comp_dict['soc_inter_variable'] = utils.SOC_INTER
        return comp_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ==============================================================================
#    S O L A R
# ==============================================================================
"""
* File name: solar.py
* Last edited: 2020-06-30
* Created by: Stefan Bruche (TU Berlin)

The solar classes SolarData, SolarThermalCollector, and PVSystem provide
functionality to model feed-in time series data for solar components (thermal
and electrical) at a certain location and with specific tilt and azimuth angles.

.. note::
   The solar classes require the availability of the Python module *pvlib*.
   The module is not provided with the standard installation of *aristopy*.
   If you want to use the solar classes, consider installing the module in your
   current environment, e.g. via: ::

   >> pip install pvlib

   For further information and an installation guide, users are referred to the
   `pvlib documentation <https://pvlib-python.readthedocs.io/en/stable/>`_.
"""
import inspect
import pandas as pd
try:
    import pvlib
    HAS_PVLIB = True
except ImportError:
    HAS_PVLIB = False


class SolarData:
    def __init__(self, ghi, dhi, latitude, longitude, altitude=0, dni=None):
        """
        Class to provide solar input data for PV or solar thermal calculations.

        The main output is accessed via function "get_plane_of_array_irradiance"
        for the POA data with specified surface tilt and azimuth values and
        "get_irradiance_dataframe" for a pandas DataFrame consisting of GHI, DHI
        and DNI values at the specified location and time index.

        :param ghi: Pandas series (with datetime index and time zone) for global
            horizontal irradiation data at the respective location.
        :param dhi: Pandas series (with datetime index and time zone) for
            diffuse horizontal irradiation data at the respective location.
        :param latitude: Latitude value (float, int) for respective location.
        :param longitude: Longitude value (float, int) for respective location.
        :param altitude: Altitude value (float, int) for respective location.
        :param dni: Pandas series (with datetime index and time zone) for direct
            normal (beam) irradiation data at the respective location.
            Is calculated from GHI and DHI if not provided here.
        """
        if not HAS_PVLIB:
            raise ImportError(
                'Module "pvlib" not found. Please consider installing it in '
                'your current environment via "pip install pvlib" if you '
                'want to use the solar class "%s".' % self.__class__.__name__)

        self.solar_position = None  # init

        self.ghi = ghi
        self.dhi = dhi
        self.dni = dni

        self.location = (latitude, longitude, altitude)

    @property
    def ghi(self):
        return self._ghi

    @ghi.setter
    def ghi(self, ghi):
        if not isinstance(ghi, pd.Series):
            raise TypeError('Input for GHI needs to be a pandas Series')
        self._ghi = ghi
        self._data_index = ghi.index
        self._tz = ghi.index.tz
        # Update the solar position for the location and the time index
        # Not during first initialization --> location is inited later
        if hasattr(self, 'location'):
            self.solar_position = self.location.get_solarposition(
                times=self._data_index)

    @property
    def data_index(self):
        return self._data_index

    @property
    def dhi(self):
        return self._dhi

    @dhi.setter
    def dhi(self, dhi):
        if not isinstance(dhi, pd.Series):
            raise TypeError('Input for DHI needs to be a pandas Series')
        if not dhi.index.equals(self._data_index):
            raise ValueError('GHI and DHI need to have the same data index')
        if dhi.index.tz != self._tz:
            raise ValueError('GHI and DHI need to have the same time zone')
        self._dhi = dhi

    @property
    def dni(self):
        return self._dni

    @dni.setter
    def dni(self, dni):
        if isinstance(dni, type(None)):
            self._dni = None
            return
        if not isinstance(dni, pd.Series):
            raise TypeError('Input for DNI needs to be a pandas Series')
        if not dni.index.equals(self._data_index):
            raise ValueError('DNI needs the same data index as GHI and DHI')
        if dni.index.tz != self._tz:
            raise ValueError('DNI needs the same time zone as GHI and DHI')
        self._dni = dni

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, location):
        if isinstance(location, pvlib.location.Location):
            self._location = location
        elif isinstance(location, tuple) and len(location) == 3:
            self._location = pvlib.location.Location(
                latitude=location[0], longitude=location[1],
                altitude=location[2], tz=self._tz)
        # Update the solar position for the location and the time index
        self.solar_position = self._location.get_solarposition(
            times=self.data_index)

    def get_dni(self):
        # If DNI already specified during initialization of SolarData => use it!
        if not isinstance(self.dni, type(None)):
            return self.dni

        # Else: Calculate DNI in two steps:
        # 1. Calculate clear sky data for the specified location by using model
        # 'simplified_solis'. The method returns a DataFrame with ghi, dni, dhi.
        # The clear sky 'dni' value is used to validate and restrict the dni
        # returned by the method "pvlib.irradiance.dni".
        clearsky = self.location.get_clearsky(
            times=self.solar_position.index, model='simplified_solis',
            solar_position=self.solar_position)

        # 2. DNI from GHI, DHI and solar positions. DNI may be unreasonably
        # high or neg. for zenith angles close to 90° (sunrise/sunset
        # transitions). Function sets them to NaN => correct to 0
        self.dni = pvlib.irradiance.dni(
            ghi=self.ghi, dhi=self.dhi,
            zenith=self.solar_position['apparent_zenith'],
            clearsky_dni=clearsky['dni'], clearsky_tolerance=1.1).fillna(0)

        return self.dni

    def get_plane_of_array_irradiance(self, surface_tilt, surface_azimuth,
                                      **kwargs):
        """
        Calculate and return the plane of array irradiance (POA).

        :param surface_tilt: tilt of the modules (0=horizontal, 90=vertical)
        :param surface_azimuth: module azimuth angle (180=facing south)
        :param kwargs: Option to specify more keyword arguments, e.g, 'albedo'
            or 'surface_type'. Search for method 'get_total_irradiance' in the
            documentation of pvlib for further information.
        :return: pandas DataFrame with POA ('poa_global', ...)
        """
        # Inspect pvlib-function for accepted arguments and filter the kwargs
        accepted_kwargs = {}
        kw_options = inspect.signature(pvlib.irradiance.get_total_irradiance)
        for key, val in kwargs.items():
            if key in kw_options.parameters.keys():
                accepted_kwargs[key] = val

        return pvlib.irradiance.get_total_irradiance(
            surface_tilt=surface_tilt, surface_azimuth=surface_azimuth,
            dni=self.get_dni(), ghi=self.ghi, dhi=self.dhi,
            solar_zenith=self.solar_position['apparent_zenith'],
            solar_azimuth=self.solar_position['azimuth'], **accepted_kwargs)

    def get_irradiance_dataframe(self):
        """
        Create and return a pandas DataFrame consisting of global (GHI) and
        diffuse horizontal (DHI) and direct normal irradiation (DNI).

        :return: pandas DataFrame with column names 'ghi', 'dhi', 'dni'
        """
        df = pd.DataFrame(index=self.ghi.index)
        df['ghi'] = self.ghi
        df['dhi'] = self.dhi
        df['dni'] = self.get_dni()
        return df


class SolarThermalCollector:
    def __init__(self, **kwargs):
        """
        Required input arguments (either while creating the class object or
        while calling function 'get_collector_heat_output'.

        * 'optical_efficiency': Opt. eff of the collector (float, int)
        * 'thermal_loss_parameter_1': Th. loss of the collector (float, int)
        * 'thermal_loss_parameter_2': Th. loss of the collector (float, int)
          |br| => See equation in: V.Quaschning, 'Regenerative Energiesysteme',
          10th edition, Hanser, 2019, p.131ff.
        * 'irradiance_data': Irradiance (POA) on collector array (pd.Series)
        * 't_ambient': Ambient temperature (float, int, pd.Series)
        * 't_collector_in': Collector inlet temperature (float, int, pd.Series)
        * 't_collector_out': Collector outlet temp. (float, int, pd.Series)
        """
        if not HAS_PVLIB:
            raise ImportError(
                'Module "pvlib" not found. Please consider installing it in '
                'your current environment via "pip install pvlib" if you '
                'want to use the solar class "%s".' % self.__class__.__name__)

        # initialize private helper variables with None
        self._all_private_vars = [
            '_optical_efficiency', '_thermal_loss_parameter_1',
            '_thermal_loss_parameter_2', '_t_ambient',
            '_t_collector_in', '_t_collector_out', '_irradiance_data']

        for var in self._all_private_vars:
            setattr(self, var, None)

        # Initialize the target values
        self.t_delta = None
        self.collector_efficiency = None
        self.collector_heat = None

        # Set the keyword arguments
        for key, val in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, val)

    @property
    def t_collector_in(self):
        return self._t_collector_in

    @t_collector_in.setter
    def t_collector_in(self, value):
        if not isinstance(value, (float, int, pd.Series)):
            raise ValueError('Expected numeric value or pandas Series.')
        self._t_collector_in = value
        self._trigger_calculation()

    @property
    def t_collector_out(self):
        return self._t_collector_out

    @t_collector_out.setter
    def t_collector_out(self, value):
        if not isinstance(value, (float, int, pd.Series)):
            raise ValueError('Expected numeric value or pandas Series.')
        self._t_collector_out = value
        self._trigger_calculation()

    @property
    def irradiance_data(self):
        return self._irradiance_data

    @irradiance_data.setter
    def irradiance_data(self, value):
        if not isinstance(value, pd.Series):
            raise ValueError('Irradiance data required as pandas Series')
        self._irradiance_data = value
        self._trigger_calculation()

    @property
    def optical_efficiency(self):
        return self._optical_efficiency

    @optical_efficiency.setter
    def optical_efficiency(self, value):
        if not isinstance(value, (int, float)) or not 0 < value <= 1:
            raise ValueError('Expected a value between 0 and 1 for opt. eff.')
        self._optical_efficiency = value
        self._trigger_calculation()

    @property
    def thermal_loss_parameter_1(self):
        return self._thermal_loss_parameter_1

    @thermal_loss_parameter_1.setter
    def thermal_loss_parameter_1(self, value):
        if not isinstance(value, (float, int)):
            raise ValueError('Expected a numeric value for the th. loss param.')
        self._thermal_loss_parameter_1 = value
        self._trigger_calculation()

    @property
    def thermal_loss_parameter_2(self):
        return self._thermal_loss_parameter_2

    @thermal_loss_parameter_2.setter
    def thermal_loss_parameter_2(self, value):
        if not isinstance(value, (float, int)):
            raise ValueError('Expected a numeric value for the th. loss param.')
        self._thermal_loss_parameter_2 = value
        self._trigger_calculation()

    @property
    def t_ambient(self):
        return self._t_ambient

    @t_ambient.setter
    def t_ambient(self, value):
        if not isinstance(value, (float, int, pd.Series)):
            raise ValueError('Expected numeric value or pandas Series.')
        self._t_ambient = value
        self._trigger_calculation()

    def _has_all_inputs(self):
        # flag to check if collector is ready for calculation of heat output
        has_all_inputs = True
        for var in self._all_private_vars:
            if getattr(self, var) is None:
                has_all_inputs = False
        return has_all_inputs

    def _matching_indices(self):
        # Get all indices and store them in temporary list:
        indices = [getattr(self, var).index for var in self._all_private_vars
                   if isinstance(getattr(self, var), pd.Series)]
        # Check that all indices are the same:
        if len(indices) > 1:
            for idx in indices[1:]:  # starting with second element
                if not idx.equals(indices[0]):  # not equal first element?
                    return False
        # list of indices has only one element or no return of False
        return True

    def _trigger_calculation(self):
        # calculation of output is triggered whenever a parameter is changed
        # but only if all required input values are available and indices match.
        if self._has_all_inputs():
            if not self._matching_indices():
                raise ValueError('The provided data indices do not match!')
            else:
                t_mean = (self.t_collector_in + self.t_collector_out) / 2
                self.t_delta = t_mean - self.t_ambient
                self.collector_efficiency = \
                    self.optical_efficiency \
                    - self.thermal_loss_parameter_1 * self.t_delta \
                    / self.irradiance_data \
                    - self.thermal_loss_parameter_2 * self.t_delta**2 \
                    / self.irradiance_data
                # Replace negative values with zeros
                self.collector_efficiency[self.collector_efficiency < 0] = 0

                self.collector_heat = \
                    self.irradiance_data * self.collector_efficiency

    def get_collector_heat_output(self, **kwargs):
        """
        Required input arguments (either while creating the class object or
        while calling function 'get_collector_heat_output'.

        * 'optical_efficiency': Opt. eff of the collector (float, int)
        * 'thermal_loss_parameter_1': Th. loss of the collector (float, int)
        * 'thermal_loss_parameter_2': Th. loss of the collector (float, int)
          |br| => See equation in: V.Quaschning, 'Regenerative Energiesysteme',
          10th edition, Hanser, 2019, p.131ff.
        * 'irradiance_data': Irradiance (POA) on collector array (pd.Series)
        * 't_ambient': Ambient temperature (float, int, pd.Series)
        * 't_collector_in': Collector inlet temperature (float, int, pd.Series)
        * 't_collector_out': Collector outlet temp. (float, int, pd.Series)

        :return: pandas Series of provided collector heat output
        """
        # Set the keyword arguments.
        # The calculation is performed while setting input arguments.
        for key, val in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, val)

        # Print a hint which required variables are still missing (if any)
        for var in self._all_private_vars:
            if getattr(self, var) is None:
                print('Missing input parameter "%s" detected' % var[1:])

        # Export the result
        return self.collector_heat


class PVSystem:
    def __init__(self, module, inverter):
        """
        PVSystem class holds a type of PV module and PV inverter.

        The main class function is "get_feedin", to get the power feedin for a
        PV plant with specified module, inverter, tilt and azimuth angle,
        location and weather data.

        :param module: Name of the module as in PVLib Database. To see the full
            database type e.g. "pvlib.pvsystem.retrieve_sam(name='cecmod')"
        :param inverter: Name of the inverter as in PVLib Database. To see the
            database type e.g. "pvlib.pvsystem.retrieve_sam(name='cecinverter')"
        """
        if not HAS_PVLIB:
            raise ImportError(
                'Module "pvlib" not found. Please consider installing it in '
                'your current environment via "pip install pvlib" if you '
                'want to use the solar class "%s".' % self.__class__.__name__)

        self.system = pvlib.pvsystem.PVSystem()  # create empty PVSystem

        self.module = module
        self.inverter = inverter

        # Initialization of default values and (private) attributes
        self.mode = 'ac'
        self._location = None
        self.mc = None  # ModelChain

    @property
    def module(self):
        return self._module

    @module.setter
    def module(self, module):
        if module in pvlib.pvsystem.retrieve_sam(name='cecmod'):
            self._module_parameters = pvlib.pvsystem.retrieve_sam(
                name='cecmod')[module]
        elif module in pvlib.pvsystem.retrieve_sam(name='sandiamod'):
            self._module_parameters = pvlib.pvsystem.retrieve_sam(
                name='sandiamod')[module]
        else:
            raise ValueError('Module %s not found in the database.' % module)
        # If not raised, do ...
        self._module = module
        self.system.module_parameters = self._module_parameters

    @property
    def module_parameters(self):
        return self._module_parameters

    @property
    def inverter(self):
        return self._inverter

    @inverter.setter
    def inverter(self, inverter):
        if inverter in pvlib.pvsystem.retrieve_sam(name='cecinverter'):
            self._inverter_parameters = pvlib.pvsystem.retrieve_sam(
                name='cecinverter')[inverter]
        elif inverter in pvlib.pvsystem.retrieve_sam(name='sandiainverter'):
            self._inverter_parameters = pvlib.pvsystem.retrieve_sam(
                name='sandiainverter')[inverter]
        elif inverter in pvlib.pvsystem.retrieve_sam(name='adrinverter'):
            self._inverter_parameters = pvlib.pvsystem.retrieve_sam(
                name='adrinverter')[inverter]
        else:
            raise ValueError('Inverter %s not found in the database' % inverter)
        # If not raised, do ...
        self._inverter = inverter
        self.system.inverter_parameters = self._inverter_parameters

    @property
    def inverter_parameters(self):
        return self._inverter_parameters

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, mode):
        if mode.lower() == 'ac' or mode.lower() == 'dc':
            self._mode = mode.lower()
        else:
            raise ValueError("Mode must either be 'ac' or 'dc'")

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, location):
        if isinstance(location, pvlib.location.Location):
            self._location = location
        else:
            raise ValueError('Please provide a valid PVLib location instance!')

    @property
    def area(self):
        """
        Get area of the PV system in :math:`m^2`

        :return: PV System area
        """
        if 'Area' in self.system.module_parameters.index:
            area_per_module = self.system.module_parameters.Area
        elif 'A_c' in self.system.module_parameters.index:
            area_per_module = self.system.module_parameters.A_c
        else:
            return None
        return (area_per_module
                * self.system.strings_per_inverter
                * self.system.modules_per_string)

    @property
    def peak_power(self):
        """
        PV system peak power [W] can be limited by the inverter or the modules
        (minimum). If DC mode is selected the inverter is not considered.

        :return: Peak power of the PV System
        """
        # I and V at the MPP have different names depending on the database
        mod_par = self.system.module_parameters
        i_mpo = mod_par.Impo if hasattr(mod_par, 'Impo') else mod_par.I_mp_ref
        v_mpo = mod_par.Vmpo if hasattr(mod_par, 'Vmpo') else mod_par.V_mp_ref

        if self.mode == "ac":
            return min(
                i_mpo * v_mpo
                * self.system.strings_per_inverter
                * self.system.modules_per_string,
                self.system.inverter_parameters.Paco)
        else:  # self.mode == "dc":
            return (i_mpo * v_mpo
                    * self.system.strings_per_inverter
                    * self.system.modules_per_string)

    def set_location(self, latitude, longitude, altitude):
        self.location = pvlib.location.Location(
            latitude=latitude, longitude=longitude, altitude=altitude)

    def get_feedin(self, weather, surface_tilt, surface_azimuth,
                   scaling=None, mode='ac', **kwargs):
        """
        :param weather: requires pandas DataFrame with at least two out of three
            column names: 'ghi', 'dhi', 'dni'. Additionally users can specify
            column names 'wind_speed' and 'temp_air' (used in calc. of losses)
        :param surface_tilt: tilt of the PV modules (0=horizontal, 90=vertical)
        :param surface_azimuth: module azimuth angle (180=facing south)
        :param scaling:
            a) None=no feed-in scaling [W],
            b) 'area'=scale feed-in to area [W/m2],
            c) 'peak_power'=scale feed-in to nominal power [-]
        :param mode:
            a) 'ac': return AC feed-in (including inverter),
            b) 'dc': return DC feed-in (excluding inverter)
        :param kwargs: Examples for kwargs are 'albedo', 'modules_per_string',
            'strings_per_inverter', 'temperature_model_parameters', ...

        :return: pandas DataFrame with POA ('poa_global', ...)
        """
        if 'location' in kwargs.keys():
            self.location = kwargs['location']
        self.mode = mode

        self.system.surface_tilt = surface_tilt
        self.system.surface_azimuth = surface_azimuth

        # Construct a ModelChain instance
        try:
            self.mc = pvlib.modelchain.ModelChain(self.system, self.location)
        except ValueError:
            # If parameters can not be found in module_parameters, try:
            self.mc = pvlib.modelchain.ModelChain(self.system, self.location,
                                                  aoi_model='no_loss',
                                                  spectral_model='no_loss')
        # Set additional keyword arguments to PV System and ModelChain
        for key, val in kwargs.items():
            if key in dir(self.system):
                setattr(self.system, key, val)
            elif key in dir(self.mc):
                setattr(self.mc, key, val)

        # Check that ghi, dhi, dni are available or at least 2 out of 3
        has_ghi = 1 if 'ghi' in [i.lower() for i in weather] else 0
        has_dhi = 1 if 'dhi' in [i.lower() for i in weather] else 0
        has_dni = 1 if 'dni' in [i.lower() for i in weather] else 0
        if has_ghi + has_dhi + has_dni < 2:
            raise ValueError('Need at least 2 out of 3 from GHI, DHI, DNI '
                             'provided in weather Dataframe.')
        elif has_ghi + has_dhi + has_dni == 2:
            self.mc.complete_irradiance(weather=weather)  # Might need pytables?

        # run the model chain
        self.mc.run_model(weather=weather)

        # Get feed-in power in AC or DC, scale it to area or peak_power if
        # required and export result:
        feedin = self.mc.ac if self.mode == 'ac' else self.mc.dc.p_mp
        # Replace negative values with zeros
        feedin[feedin < 0] = 0

        if scaling is None:
            return feedin
        elif isinstance(scaling, str) and scaling.lower() == 'area':
            return feedin / float(self.area)
        elif isinstance(scaling, str) and scaling.lower() == 'peak_power':
            return feedin / float(self.peak_power)
        else:
            raise ValueError('Valid input for parameters "scaling" are None, '
                             '"area" or "peak_power"')


if __name__ == '__main__':

    # Set time index (incl. time zone) and sample data for global (GHI) and
    # diffuse horizontal irradiation (DHI) [W/m²]:
    idx = pd.date_range(start='2018-01-01 12:00', periods=12, freq='M', tz='UTC')
    ghi = pd.Series([0, 50, 150, 400, 600, 900, 1000, 600, 400, 150, 50, 0], idx)
    dhi = pd.Series([0, 30, 100, 200, 300, 400, 0, 100, 50, 10, 0, 0], idx)

    # SolarData:
    # ----------
    solar = SolarData(ghi=ghi, dhi=dhi, latitude=52.3822,
                      longitude=13.0622, altitude=81)

    # Solar thermal collector:
    # ------------------------
    poa = solar.get_plane_of_array_irradiance(
        surface_tilt=45, surface_azimuth=180)
    solar_coll = SolarThermalCollector(
        optical_efficiency=0.73, thermal_loss_parameter_1=1.7,
        thermal_loss_parameter_2=0.016, irradiance_data=poa['poa_global'],
        t_ambient=20, t_collector_in=20, t_collector_out=40)
    heat_out = solar_coll.get_collector_heat_output()

    # PV-System:
    # ----------
    # If available: append wind and temperature data to the DataFrame here
    weather = solar.get_irradiance_dataframe()
    # Create a PV System (consisting of a module and an inverter)
    pv_sys = PVSystem(module='Canadian_Solar_CS5P_220M___2009_',
                      inverter='ABB__MICRO_0_25_I_OUTD_US_208__208V_')
    # Calculate the feed-in of the PV system for specified weather conditions at
    # a site and a collector position.
    temp = pvlib.temperature.TEMPERATURE_MODEL_PARAMETERS['sapm'][
        'open_rack_glass_glass']
    feedin = pv_sys.get_feedin(weather=weather, location=solar.location,
                               surface_tilt=25, surface_azimuth=180,
                               temperature_model_parameters=temp)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ==============================================================================
#    E N E R G Y S Y S T E M
# ==============================================================================
"""
* File name: energySystem.py
* Last edited: 2020-06-14
* Created by: Stefan Bruche (TU Berlin)

The EnergySystem class is aristopy's main model container. An instance of
the EnergySystem class holds the modeled components, the overall pyomo
model and the results of the optimization.
The EnergySystem class provides features to built and solve the
optimization problem, manipulate the associated component models, and
process the results of the optimization. The implemented class methods are:

* :meth:`cluster <aristopy.energySystem.EnergySystem.cluster>`:
  Perform clustering of the implemented time series data

* :meth:`declare_model <aristopy.energySystem.EnergySystem.declare_model>`:
  Declare the pyomo optimization model

* :meth:`optimize <aristopy.energySystem.EnergySystem.optimize>`:
  Call the main optimization routine

* :meth:`relax_integrality
  <aristopy.energySystem.EnergySystem.relax_integrality>`:
  Relax the integrality of binary variables

* :meth:`edit_component_variables
  <aristopy.energySystem.EnergySystem.edit_component_variables>`:
  Edit properties of component variables, e.g., change bounds or domains

* :meth:`reset_component_variables
  <aristopy.energySystem.EnergySystem.reset_component_variables>`:
  Reset component variables after applying changes, e.g., relaxation

* :meth:`export_component_configuration
  <aristopy.energySystem.EnergySystem.export_component_configuration>`,
  :meth:`import_component_configuration
  <aristopy.energySystem.EnergySystem.import_component_configuration>`,:
  Export and import configurations, i.e. component existences and capacities

* :meth:`add_design_integer_cut_constraint
  <aristopy.energySystem.EnergySystem.add_design_integer_cut_constraint>`:
  Create integer-cut-constraints to exclude the current design solution
  from the solution space and enforce a new design in subsequent model runs

* :meth:`add_variable <aristopy.energySystem.EnergySystem.add_variable>`,
  :meth:`add_constraint <aristopy.energySystem.EnergySystem.add_constraint>`,
  :meth:`add_objective_function_contribution
  <aristopy.energySystem.EnergySystem.add_objective_function_contribution>`:
  Add variables, constraints and objective function contributions directly
  to the main pyomo model, outside of the component declaration
"""
import os
import time
import json
from collections import OrderedDict

import pandas as pd
import pyomo.environ as pyo
import pyomo.network as network
import pyomo.opt as opt
from tsam.timeseriesaggregation import TimeSeriesAggregation

from aristopy import utils, logger


class EnergySystem:
    def __init__(self, number_of_time_steps=8760, hours_per_time_step=1,
                 interest_rate=0.05, economic_lifetime=20, logging=None):
        """
        Initialize an instance of the EnergySystem class.

        :param number_of_time_steps: Number of considered time steps for
            modeling the dispatch problem. With "hours_per_time_step" the
            share of the modeled year can be calculated. In this way, the cost
            of each time step is scaled and included in the objective function.
            |br| *Default: 8760*
        :type number_of_time_steps: int (>0)

        :param hours_per_time_step: Number of hours per modeled time step.
            |br| *Default: 1*
        :type hours_per_time_step: int (>0)

        :param interest_rate: Value to calculate the present value factor of a
            cost rate that occurs in the future.
            |br| *Default: 0.05 (corresponds to 5%)*
        :type interest_rate: float, int (>=0)

        :param economic_lifetime: Years to consider for calculating the net
            present value of an investment with annual incoming and outgoing
            cash flows.
            |br| *Default: 20*
        :type economic_lifetime:  int (>0)

        :param logging: Specify the behavior of the logging by setting an own
            Logger class instance. User can decide where to log (file/console)
            and what to log (see description of aristopy "Logger").
            |br| *Default: None (display minimal logging in the console)*
        :type logging: None or instance of aristopy's "Logger" class
        """
        # Check user input:
        utils.check_energy_system_input(
            number_of_time_steps, hours_per_time_step, interest_rate,
            economic_lifetime, logging)

        # **********************************************************************
        #   Logging
        # **********************************************************************
        # If no logger instance is passed to the "logging" keyword a default
        # logger is initialized. This will only display errors on the console.
        # Otherwise the passed logger instance is used and a logger for the
        # instance of the energy system class is initialized on "self.log"
        if logging is None:
            self.logger = logger.Logger(default_log_handler='stream',
                                        default_log_level='ERROR')
        else:
            self.logger = logging
        self.log = self.logger.get_logger(self)

        # **********************************************************************
        #   Time and clustering
        # **********************************************************************
        self.number_of_time_steps = number_of_time_steps
        self.hours_per_time_step = hours_per_time_step
        self.number_of_years = number_of_time_steps * hours_per_time_step/8760.0

        # Initialization: Overwritten if 'cluster' function is called
        self.periods = [0]
        self.periods_order = [0]
        self.period_occurrences = [1]
        self.number_of_time_steps_per_period = number_of_time_steps
        self.inter_period_time_steps = [0, 1]  # one before & after only period

        # Flag 'is_data_clustered' indicates if the function 'cluster' has been
        # called before. The flag is reset to False if new components are added.
        self.is_data_clustered = False
        # 'typical_periods' is altered by function 'cluster' to an array ranging
        # from 0 to number_of_typical_periods-1.
        self.typical_periods = [0]

        # Stores the instance of the time series aggregation (if performed)
        self.tsa_instance = None

        # **********************************************************************
        #   Economics
        # **********************************************************************
        # The economic criterion net present value represents the objective
        # function value to be maximized. Hence, a present value factor (pvf) is
        # required to calculate the present value of an annuity. The global
        # parameters interest rate and economic lifetime of the energy system
        # investment are used to this end.
        self.pvf = sum(1 / (1 + interest_rate)**n
                       for n in range(1, economic_lifetime+1))

        # **********************************************************************
        #   Optimization
        # **********************************************************************
        # The parameter 'model' holds the pyomo ConcreteModel instance with
        # sets, parameters, variables, constraints and the objective function.
        # It is None during initialization and changed when the functions
        # 'optimize', or 'declare_model' are called.
        # Before the model instance is optimized, a solver instance is assigned
        # to the "solver" attribute. It also stores basic results of the run.
        # The "is_model_declared" flag indicates if the model instance is
        # already declared.
        # The "is_persistent_model_declared" flag states if the model has been
        # declared and assigned to a persistent solver instance.
        self.model = None
        self.run_info = {'solver_name': '',
                         'time_limit': None, 'optimization_specs': '',
                         'model_build_time': 0, 'model_solve_time': 0,
                         'upper_bound': 0, 'lower_bound': 0, 'sense': '',
                         'solver_status': '', 'termination_condition': ''}
        self.solver = None
        self.is_model_declared = False
        self.is_persistent_model_declared = False

        # **********************************************************************
        #   Components
        # **********************************************************************
        # 'components' is a dictionary {component name: component object itself}
        # in which all components of the EnergySystem instance are stored.
        # The pyomo block model object (stored variables and constraints) of a
        # component instance can be accessed via its "block" attribute.
        self.components = {}

        # The 'component_connections' is a dict that stores the connections of
        # the component instances of the energy system model. It is formed from
        # the specified inlets and outlets and the connecting commodity:
        # {arc_name: [source instance, destination instance, commodity_name]}
        self.component_connections = {}

        # 'component_configuration' is a pandas Dataframe to store basic
        # information about the availability and capacity of the modelled
        # components. It is used to export / import the configuration results.
        self.component_configuration = pd.DataFrame(
            index=[utils.BI_EX, utils.BI_MODULE_EX, utils.CAP])

        # DataFrames and dictionaries to store additionally added pyomo objects
        # (variables and constraints) and objective function contributions.
        self.added_constraints = pd.DataFrame(index=['has_time_set',
                                                     'alternative_set', 'rule'])
        self.added_variables = pd.DataFrame(index=['domain', 'has_time_set',
                                                   'alternative_set', 'init',
                                                   'ub', 'lb', 'pyomo'])
        self.added_objective_function_contributions = {}
        self.added_obj_contributions_results = {}

        self.log.info('Initializing EnergySystem completed!')

    def __repr__(self):
        # Define class format for printing and logging
        return '<EnSys: "id=%s..%s">' % (hex(id(self))[:3], hex(id(self))[-3:])

    def add_variable(self, var):
        """
        Function to manually add pyomo variables to the main pyomo model
        (ConcreteModel: model) of the energy system instance via instances
        of aristopy's Var class. The attributes of the variables are stored in
        DataFrame "added_variables" and later initialized during the call of
        function 'optimize', or 'declare_model'.

        :param var: Instances of aristopy's Var class (single or in list)
        """
        self.log.info('Call of function "add_variable"')

        # Check the correctness of the user input
        var_list = utils.check_add_vars_input(var)

        for v in var_list:
            # Wrap everything up in a pandas Series
            series = pd.Series({'has_time_set': v.has_time_set,
                                'alternative_set': v.alternative_set,
                                'domain': v.domain, 'init': v.init,
                                'ub': v.ub, 'lb': v.lb, 'pyomo': None})
            # Add the Series with new name to DataFrame "added_variables"
            self.added_variables[v.name] = series

    def add_constraint(self, rule, name=None, has_time_set=True,
                       alternative_set=None):
        """
        Function to manually add constraints to the main pyomo model after the
        instance has been created. The attributes are stored in the DataFrame
        'added_constraints' and later initialized during the call of function
        'optimize', or 'declare_model'.

        :param rule: A Python function that specifies the constraint with a
            equality or inequality expression. The rule must hold at least
            two arguments: First the energy system instance it is added to (in
            most cases: self), second the ConcreteModel of the instance (model).
            Additional arguments represent sets (e.g., time).
        :type rule: function

        :param name: Name (identifier) of the added constraint. The rule name is
            used if no name is specified.
            |br| *Default: None*
        :type name: str

        :param has_time_set: Is True if the time set of the energy system model
            is also a set of the added constraint.
            |br| *Default: True*
        :type has_time_set: bool

        :param alternative_set: Alternative constraint sets can be added here
            via iterable Python objects (e.g. list).
            |br| *Default: None*
        """
        self.log.info('Call of function "add_constraint"')

        # Check the correctness of the user input
        utils.check_add_constraint(rule, name, has_time_set, alternative_set)
        # The rule name is used as constraint identifier if no name is given
        if name is None:
            name = rule.__name__
        # Put everything together in a pandas Series
        series = pd.Series({'has_time_set': has_time_set,
                            'alternative_set': alternative_set,
                            'rule': rule})
        # Add the Series to the DataFrame "added_constraints"
        self.added_constraints[name] = series

    def add_objective_function_contribution(self, rule, name=None):
        """
        Additional objective function contributions can be added with this
        method. The method requires a Python function input that takes the main
        pyomo model (ConcreteModel: model) and returns a single (scalar) value.

        :param rule: A Python function returning a scalar value which is added
            to the objective function of the model instance. The rule must hold
            exactly two arguments: The energy system instance it is added to (in
            most cases: self), second the ConcreteModel of the instance (model).
        :type rule: function

        :param name: Name (identifier) of the added objective function
            contribution. The rule name is used if no name is specified.
            |br| *Default: None*
        :type name: str
        """
        self.log.info('Call of function "add_objective_function_contribution"')

        # Check the input:
        assert isinstance(name, (str, type(None))), '"name" should be a string!'
        if not callable(rule):
            raise TypeError('The "rule" needs to hold a callable object!')
        if name is None:
            name = rule.__name__
        # Add the rule and the name to a dictionary of the EnergySystem instance
        self.added_objective_function_contributions[name] = rule

    def cluster(self, number_of_typical_periods=4,
                number_of_time_steps_per_period=24,
                cluster_method='hierarchical',
                **kwargs):
        """
        Method for the aggregation and clustering of time series data. First,
        the time series data and their respective weights are collected from all
        components and split into pieces with equal length of
        'number_of_time_steps_per_period'.
        Subsequently, a clustering method is called and each period is assigned
        to one of 'number_of_typical_periods' typical periods. The clustered
        data is later stored in the components.
        The package `tsam <https://github.com/FZJ-IEK3-VSA/tsam>`_ (time series
        aggregation module) is used to perform the clustering.
        The clustering algorithm can be controlled by adding required keyword
        arguments (using 'kwargs' parameter). To learn more about tsam and
        possible keyword arguments see the package `documentation
        <https://tsam.readthedocs.io/en/latest/index.html>`_.

        :param number_of_typical_periods: Number of typical periods to be
            clustered. |br| *Default: 4*
        :type number_of_typical_periods: int (>0)

        :param number_of_time_steps_per_period: Number of time steps per period
            |br| *Default: 24*
        :type number_of_time_steps_per_period: int (>0)

        :param cluster_method: Name of the applied clustering method (e.g.,
            'k_means'). See the tsam documentation for all possible options.
            |br| *Default: 'hierarchical'*
        :type cluster_method: str
        """
        # Check input arguments
        utils.check_cluster_input(number_of_typical_periods,
                                  number_of_time_steps_per_period,
                                  self.number_of_time_steps)

        time_start = time.time()
        self.log.info('Start clustering with %s typical periods and %s time '
                      'steps per period.' % (number_of_typical_periods,
                                             number_of_time_steps_per_period))

        # Get time series data and their respective weights from all components
        # and collect them in two dictionaries
        time_series_data, time_series_weights = {}, {}
        for comp in self.components.values():
            if comp.number_in_group == 1:  # Add only once per group
                data, weights = comp.get_time_series_data_for_aggregation()
                time_series_data.update(data)
                time_series_weights.update(weights)

        # Convert data dictionary to pandas DataFrame
        time_series_data = pd.DataFrame.from_dict(time_series_data)
        # Specific index is not relevant, but tsam requires a uniform index
        time_series_data.index = \
            pd.date_range('2050-01-01 00:30:00',
                          periods=self.number_of_time_steps,
                          freq=(str(self.hours_per_time_step) + 'H'),
                          tz='Europe/Berlin')
        # Reindex axis for reproducibility of TimeSeriesAggregation results
        time_series_data = time_series_data.reindex(
            sorted(time_series_data.columns), axis=1)

        # Set up instance of tsam's TimeSeriesAggregation class and cluster data
        self.tsa_instance = TimeSeriesAggregation(
            timeSeries=time_series_data,
            noTypicalPeriods=number_of_typical_periods,
            hoursPerPeriod=
            number_of_time_steps_per_period * self.hours_per_time_step,
            clusterMethod=cluster_method,
            weightDict=time_series_weights,
            **kwargs)

        # Store clustered time series data in the components
        data = pd.DataFrame.from_dict(self.tsa_instance.clusterPeriodDict)
        for comp in self.components.values():
            comp.set_aggregated_time_series_data(data)

        self.typical_periods = self.tsa_instance.clusterPeriodIdx
        self.number_of_time_steps_per_period = number_of_time_steps_per_period

        self.periods = list(range(int(
            self.number_of_time_steps / number_of_time_steps_per_period)))
        self.inter_period_time_steps = list(range(int(
            self.number_of_time_steps / number_of_time_steps_per_period) + 1))

        self.periods_order = self.tsa_instance.clusterOrder
        self.period_occurrences = [
            (self.periods_order == tp).sum() for tp in self.typical_periods]

        # Set cluster flag to True
        self.is_data_clustered = True
        self.log.info("    Time required for clustering: %.2f sec"
                      % (time.time() - time_start))
        # Debugging:
        self.log.debug('typical_periods: %s' % self.typical_periods)
        self.log.debug('periods: %s' % self.periods)
        self.log.debug('inter_period_time_steps: %s'
                       % self.inter_period_time_steps)
        self.log.debug('periods_order: %s' % self.periods_order)
        self.log.debug('period_occurrences: %s' % self.period_occurrences)

    def declare_time_sets(self, model, use_clustered_data):
        """
        Initialize time parameters and four different time sets.

        The "time_set" represents the general set. The index holds tuples of
        periods and time steps inside of these periods. In case the optimization
        is performed without time series aggregation, the set runs from [(0,0),
        (0,1), ..to.., (0,number_of_time_steps-1)].
        Otherwise: [(0,0), ..., (0,number_of_time_steps_per_period-1), (1,0),
        ..., (number_of_typical_periods-1, number_of_time_steps_per_period-1)].
        |br| The set "intra_period_time_set" holds tuples of periods and points
        in time before, after or between regular time steps inside of a period.
        Hence, the second value runs from 0 to "number_of_time_steps" (without
        aggregation) or "number_of_time_steps_per_period" respectively. |br|
        The third set "intra_period_time_set" is one-dimensional and ranges from
        0 to the overall number of periods plus 1. So if no aggregation is used
        it has only two entries [0, 1]. Otherwise is is ranging from 0 to
        (1 + number_of_time_steps / number_of_time_steps_per_period). |br|
        The "typical_periods_set" is a set ranging from 0 to the number of
        typical periods. If no aggregation is used it only holds 0. |br|

        E.g.: Case with 2 periods and 3 time steps per period |br|
        1) ______(0,0)___(0,1)___(0,2)___________(1,0)___(1,1)___(1,2)_____ |br|
        2) __(0,0)___(0,1)___(0,2)___(0,3)___(1,0)___(1,1)___(1,2)___(1,3)_ |br|
        3) 0___________________________    1    __________________________2 |br|
        4) ______________  0  ___________________________  1  _____________ |br|
        1) time_set, 2) intra_period_time_steps_set,
        3) inter_period_time_steps_set, 4) typical_periods_set |br|

        :param model: Pyomo model instance holding sets, variables, constraints
            and the objective function.
        :type model: pyomo ConcreteModel

        :param use_clustered_data: Declare optimization model with original
            (full scale) time series (=> False) or clustered data (=> True).
        :type use_clustered_data: bool
        """
        self.log.info('    Declare time sets')

        # Set the time series data for all modelled components
        for comp in self.components.values():
            comp.set_time_series_data(use_clustered_data)

        # Reset time-relevant attributes and initialize sets if full scale
        # problem is considered.
        if not use_clustered_data:
            self.periods = [0]
            self.periods_order = [0]
            self.period_occurrences = [1]
            self.number_of_time_steps_per_period = self.number_of_time_steps
            self.inter_period_time_steps = [0, 1]
            self.typical_periods = [0]

            # Define sets: Only period "0" exists
            def init_time_set(m):
                return ((0, t) for t in range(self.number_of_time_steps))

            def init_intra_period_time_set(m):
                return ((0, t) for t in range(self.number_of_time_steps + 1))

        else:
            self.log.info('    Aggregated time series data detected (number of '
                          'typical periods: %s, number of time steps per '
                          'period: %s' % (len(self.typical_periods),
                                          self.number_of_time_steps_per_period))

            # Define sets
            def init_time_set(m):
                return ((p, t) for p in self.typical_periods
                        for t in range(self.number_of_time_steps_per_period))

            def init_intra_period_time_set(m):
                return ((p, t) for p in self.typical_periods
                        for t in range(self.number_of_time_steps_per_period+1))

        # Initialize the four time sets:
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # E.g.: Case with 2 periods and 3 time steps per period
        # 1) ______(0,0)___(0,1)___(0,2)___________(1,0)___(1,1)___(1,2)______
        # 2) __(0,0)___(0,1)___(0,2)___(0,3)___(1,0)___(1,1)___(1,2)___(1,3)__
        # 3) 0_______________________________1_______________________________2
        # 4) ________________0_______________________________1________________
        # 1) time_set, 2) intra_period_time_steps_set,
        # 3) inter_period_time_steps_set, 4) typical_periods_set

        model.time_set = pyo.Set(
            dimen=2, initialize=init_time_set, ordered=True)

        model.intra_period_time_set = pyo.Set(
            dimen=2, initialize=init_intra_period_time_set, ordered=True)

        model.inter_period_time_set = pyo.Set(
            initialize=self.inter_period_time_steps, ordered=True)

        model.typical_periods_set = pyo.Set(
            initialize=self.typical_periods, ordered=True)

    def declare_objective(self, model):
        """
        Method to declare the objective function of the optimization problem.
        The objective function contributions (CAPEX and OPEX) are calculated in
        the components itself, and collected and summarized by the
        'declare_objective' function of the EnergySystem instance.
        The objective function of the optimization is the maximization of the
        net present value.

        :param model: Pyomo model instance holding sets, variables, constraints
            and the objective function.
        :type model: pyomo ConcreteModel
        """
        self.log.info('    Declare objective function')

        def summarize_extra_objective_function_contributions():
            for key, val in self.added_objective_function_contributions.items():
                obj_rule = getattr(self, key)  # get the function
                obj_value = obj_rule(self.model)  # get returned value
                # Add the values to the results dict (used to export results)
                self.added_obj_contributions_results[key] = obj_value
            return sum(self.added_obj_contributions_results.values())

        def objective(m):
            NPV = sum(comp.get_objective_function_contribution(self, model)
                      for comp in self.components.values()) \
                  + summarize_extra_objective_function_contributions()
            return NPV
        model.Obj = pyo.Objective(rule=objective, sense=pyo.maximize)

    def declare_model(self, use_clustered_data=False,
                      declare_persistent=False,
                      persistent_solver='gurobi_persistent'):
        """
        Declare the pyomo optimization model of the EnergySystem instance.
        First, all component connections are established by using their input
        and output specifications and the plausibility of connections is
        validated. Then, the ConcreteModel instance is created and time sets,
        component model blocks, variables, ports, constraints, arcs, and the
        objective function are added.

        :param use_clustered_data: Declare optimization model with original
            (full scale) time series (=> False) or clustered data (=> True).
            |br| *Default: False*
        :type use_clustered_data: bool

        :param declare_persistent: States if a persistent model instance should
            be formed. In this case, after model declaration a persistent solver
            instance is created and the declared model instance is assigned.
            |br| *Default: False*
        :type declare_persistent: bool

        :param persistent_solver: Name of the persistent solver to be used.
            Possible options are "gurobi_persistent" and "cplex_persistent".
            Is ignored if keyword "declare_persistent" is False.
            |br| *Default: 'gurobi_persistent'*
        :type persistent_solver: str
        """
        self.log.info('Declare optimization model with '
                      'use_clustered_data=%s and declare_persistent=%s'
                      % (use_clustered_data, declare_persistent))

        time_start = time.time()

        # Check inputs:
        persistent_solver = persistent_solver.lower()
        utils.check_declare_model_input(
            use_clustered_data, self.is_data_clustered,
            declare_persistent, persistent_solver)

        # Initialize flags of model status
        self.is_model_declared = False
        self.is_persistent_model_declared = False

        # Reset time series aggregation instance for model w/o clustered data
        if not use_clustered_data:
            self.tsa_instance = None

        # ######################################################################
        #   C O M P O N E N T   C O N N E C T I O N S
        # ######################################################################
        # Helper functions:
        # -----------------
        def component_not_found(c_name):
            return('\nThe component "{}" or its instances cannot be found in '
                   'the EnergySystem model.\nThe components considered are: {}'
                   '\n Please check the names of your components and inlet and '
                   'outlet specs!'.format(c_name, list(self.components)))

        def set_connection(source, destination, commodity):
            """
            Add connections (arcs) to the "component_connections" dict.
            Update the "var_connections" dictionaries in the components:
            {'variable name': 'connected arc name'}
            "var_connections" is used for plotting purposes.
            """
            # Add a new connection to "component_connections" if not available
            arc_name = 'arc-{}-{}-{}'.format(
                source.name, destination.name, commodity)

            if arc_name not in self.component_connections:
                self.component_connections[arc_name] = [
                    source, destination, commodity]

            # Get the commodity variables from comp. and raise if not available
            if source.outlet_commod_and_var_names.get(commodity) is None:
                raise ValueError('Commodity "%s" not found at outlet of "%s"'
                                 % (commodity, source.name))
            source_var = source.outlet_commod_and_var_names[commodity]
            if destination.inlet_commod_and_var_names.get(commodity) is None:
                raise ValueError('Commodity "%s" not found at inlet of "%s"'
                                 % (commodity, destination.name))
            destination_var = destination.inlet_commod_and_var_names[commodity]

            # Update dictionary "var_connections" in the source component
            if source.var_connections.get(source_var) is None:
                source.var_connections[source_var] = [arc_name]
            else:
                arc_names = source.var_connections[source_var]
                if arc_name not in arc_names:
                    arc_names.append(arc_name)
                    source.var_connections[source_var] = arc_names

            # Update dictionary "var_connections" in the destination component
            if destination.var_connections.get(destination_var) is None:
                destination.var_connections[destination_var] = [arc_name]
            else:
                arc_names = destination.var_connections[destination_var]
                if arc_name not in arc_names:
                    arc_names.append(arc_name)
                    destination.var_connections[destination_var] = arc_names

        # **********************************************************************
        #   Establish component connections from "inlet" and "outlet"
        # **********************************************************************
        for comp_name, comp in self.components.items():
            # Loop over inlets ...
            for flow in comp.inlet:
                if flow.link is not None:
                    # loop again over all components:
                    found_src = False
                    for c_src in self.components.values():
                        if c_src.group_name == flow.link:
                            found_src = True  # found source
                            set_connection(c_src, comp, flow.commodity)
                    if not found_src:
                        raise ValueError(component_not_found(flow.link))
            # Loop over outlets ...
            for flow in comp.outlet:
                if flow.link is not None:
                    # loop again over all components:
                    found_dest = False
                    for c_dest in self.components.values():
                        if c_dest.group_name == flow.link:
                            found_dest = True  # found destination
                            set_connection(comp, c_dest, flow.commodity)
                    if not found_dest:
                        raise ValueError(component_not_found(flow.link))

        # **********************************************************************
        #   Check connections
        # **********************************************************************
        for name, comp in self.components.items():

            # Check that non-conversion comp. have only one commodity at ports!
            if comp.__class__.__name__ != 'Conversion' and \
                    len(comp.commodities) > 1:
                raise ValueError('Commodity error in "%s". Found more than one '
                                 'commodity: "%s"' % (name, comp.commodities))

            # Check that number of inlets and outlets is correct:
            nbr_of_inlets = len(comp.inlet_commod_and_var_names)
            nbr_of_outlets = len(comp.outlet_commod_and_var_names)
            # Conversion component have at least one inlet and one outlet
            if comp.__class__.__name__ == 'Conversion' \
                    and (nbr_of_inlets < 1 or nbr_of_outlets < 1):
                raise ValueError('"%s" needs at least one inlet and one outlet!'
                                 % name)
            # Source, Storage, Bus components need exactly one outlet
            if comp.__class__.__name__ in ['Source', 'Storage', 'Bus'] and \
                    nbr_of_outlets != 1:
                raise ValueError('"%s" needs one outlet, but "%s" were found!'
                                 % (name, nbr_of_outlets))
            # Sink, Storage, Bus components need exactly one inlet
            if comp.__class__.__name__ in ['Sink', 'Storage', 'Bus'] and \
                    nbr_of_inlets != 1:
                raise ValueError('"%s" needs one inlet, but "%s" were found!'
                                 % (name, nbr_of_inlets))
            # Sinks don't have outlets
            if comp.__class__.__name__ == 'Sink' and nbr_of_outlets != 0:
                raise ValueError('Sink "%s" cannot have outlets!' % name)
            # Sources don't have inlets
            if comp.__class__.__name__ == 'Source' and nbr_of_inlets != 0:
                raise ValueError('Source "%s" cannot have inlets!' % name)

            # Raise if component has inlet or outlet commodity without
            # connection to an arc
            for commod, var in comp.inlet_commod_and_var_names.items():
                if comp.var_connections.get(var) is None:
                    raise ValueError('Inlet commodity "%s" of component "%s" is'
                                     ' unconnected!' % (commod, comp.name))
            for commod, var in comp.outlet_commod_and_var_names.items():
                if comp.var_connections.get(var) is None:
                    raise ValueError('Outlet commodity "%s" of component "%s" '
                                     'is unconnected!' % (commod, comp.name))

        # **********************************************************************
        #   Check basic variable
        # **********************************************************************
        # Usually basic variables are referring to inlet or outlet commodities
        # which are time-related. Scalar variables can also be selected under
        # certain circumstances. Check permissibility of parameter combinations:
        for name, comp in self.components.items():
            # Fine, if basic variable has a time set (--> go to next comp.)
            if comp.variables[comp.basic_variable]['has_time_set']:
                continue
            # Conversion w/o 'opex_operation' and w/o binary operation var.: OK
            elif comp.__class__.__name__ == 'Conversion' and \
                    comp.opex_operation == 0 and not comp.has_bi_op:
                continue
            # Sink or Source components w/o 'opex_operation', w/o binary
            # operation var. and w/o commodity_rates, _cost or _rev. are also OK
            elif comp.__class__.__name__ in ['Sink', 'Source'] and \
                    comp.opex_operation == 0 and not comp.has_bi_op and \
                    comp.op_rate_min is None and comp.op_rate_max is None and \
                    comp.op_rate_fix is None and comp.commodity_cost == 0 and \
                    comp.commodity_cost_time_series is None and \
                    comp.commodity_revenues == 0 and \
                    comp.commodity_revenues_time_series is None:
                continue
            else:
                raise Exception(
                    'The basic variable of component "%s" works without a time '
                    'set. This functionality is only allowed for Sink, Source ' 
                    'and Conversion components w/o binary operation variables, ' 
                    'opex_operation, commodity_rates, _cost, and _revs!' % name)

        # **********************************************************************
        #   Add user expressions
        # **********************************************************************
        # Fill the "user_expressions_dict" with data from comp. initialization
        # (old data / variables are overwritten in 'comp.user_expressions_dict')
        for comp in self.components.values():
            for expr in comp.user_expressions:
                # Create a name for the expression by removing all spaces
                expr_name = expr.replace(' ', '')
                # Disassemble expr into its parts -> e.g. ['Q', '>=', '100']
                expr_pieces = utils.disassemble_user_expression(expr)
                # Append name and list of pieces to the 'user_expressions_dict'
                comp.user_expressions_dict[expr_name] = expr_pieces

        # ######################################################################
        #   P Y O M O   M O D E L
        # ######################################################################
        # Create a pyomo ConcreteModel instance to store the sets, variables,
        # constraints, objective function, etc.
        self.model = pyo.ConcreteModel()
        # Duals are stored with the name 'dual' in the model instance (might be
        # interesting for analyzing optimization results of linear models).
        self.model.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)

        # **********************************************************************
        #   Declare time sets
        # **********************************************************************
        self.declare_time_sets(self.model, use_clustered_data)

        # **********************************************************************
        #   Declare component blocks, variables, ports and constraints
        # **********************************************************************
        for comp in self.components.values():
            comp.declare_component_model_block(model=self.model)
            comp.declare_component_variables(model=self.model)
            comp.declare_component_ports()
            comp.declare_component_user_constraints(model=self.model)
            comp.declare_component_constraints(ensys=self, model=self.model)

        # **********************************************************************
        #   Declare arcs
        # **********************************************************************
        for arc_name, connection in self.component_connections.items():
            # Get the connected component class instances and the commodity name
            src, dest, commod = connection[0], connection[1], connection[2]
            # Get the ports in the component model blocks for each variable
            outlet = getattr(src.block, 'outlet_' + commod)
            inlet = getattr(dest.block, 'inlet_' + commod)
            # Create an arc to connect two ports
            setattr(self.model, arc_name, network.Arc(src=outlet, dest=inlet))

        # Call model transformation factory: Expand the arcs
        pyo.TransformationFactory("network.expand_arcs").apply_to(self.model)

        # Starting from version 5.6.9 the variable domains (and bounds) for
        # newly created variables on Arcs are not inherited anymore. The default
        # is 'Reals' and we need to overwrite this to 'NonNegativeReals'.
        # https://github.com/Pyomo/pyomo/pull/1186
        for arc_name, connection in self.component_connections.items():
            # Get the commodity variable from the expanded Block and set domain.
            arc_block = getattr(self.model, arc_name + '_expanded')
            # There is only a new commodity variable on the arc if the variables
            # are split (not for direct connection of single outlet and inlet).
            if hasattr(arc_block, connection[2]):
                arc_var = getattr(arc_block, connection[2])
                arc_var.domain = pyo.NonNegativeReals

        # Workaround for Pyomo-Versions before 5.6.9:
        # Get all constraints that end with '_split' and deactivate them
        # --> Workaround for Port.Extensive function with indexed variables
        # See: https://groups.google.com/forum/#!topic/pyomo-forum/LaoKMhyu9pA
        # for c in model.component_objects(pyomo.Constraint, active=True):
        #     if c.name.endswith('_split'):
        #         c.deactivate()

        # **********************************************************************
        #   Declare additional pyomo objects, added by:
        #       * add_variable
        #       * add_constraint
        #       * add_objective_function_contribution
        # **********************************************************************
        # Additional variables:
        # ~~~~~~~~~~~~~~~~~~~~~
        for var_name in self.added_variables:
            # Get the pandas series from DataFrame and the variable specs
            var_dict = self.added_variables[var_name]
            domain = getattr(pyo, var_dict['domain'])
            bounds = (var_dict['lb'], var_dict['ub'])
            init = var_dict['init']
            # Differentiation between variables with and without time_set
            if var_dict['has_time_set']:
                setattr(self.model, var_name, pyo.Var(
                    self.model.time_set, domain=domain, bounds=bounds,
                    initialize=init))
            elif var_dict['alternative_set'] is not None:
                setattr(self.model, var_name, pyo.Var(
                    var_dict['alternative_set'], domain=domain,
                    bounds=bounds, initialize=init))
            else:
                setattr(self.model, var_name, pyo.Var(
                    domain=domain, bounds=bounds, initialize=init))
            # Store variable in self.added_variables[var_name]['pyomo']
            pyomo_var = getattr(self.model, var_name)
            self.added_variables[var_name]['pyomo'] = pyomo_var

        # Additional constraints:
        # ~~~~~~~~~~~~~~~~~~~~~~~
        for con_name in self.added_constraints:
            con_dict = self.added_constraints[con_name]
            rule = con_dict['rule']
            # https://stackoverflow.com/questions/972/adding-a-method-to-an-existing-object-instance
            setattr(self, con_name, rule.__get__(self))
            con = getattr(self, con_name)
            # Differentiation between variables with and without time_set
            if con_dict['has_time_set']:
                setattr(self.model, con_name, pyo.Constraint(
                    self.model.time_set, rule=con))
            elif con_dict['alternative_set'] is not None:
                setattr(self.model, con_name, pyo.Constraint(
                    con_dict['alternative_set'], rule=con))
            else:
                setattr(self.model, con_name, pyo.Constraint(rule=con))

        # Additional objective function contributions:
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        for name, rule in self.added_objective_function_contributions.items():
            setattr(self, name, rule.__get__(self))

        # **********************************************************************
        #   Declare empty constraint list for design integer cuts
        # **********************************************************************
        # Attribute "integer_cut_constraints" in the ConcreteModel is
        # initialized as an empty pyomo ConstraintList while declaring the
        # optimization problem. Integer cut constraints can be added with
        # function "add_design_integer_cut_constraint".
        self.model.integer_cut_constraints = pyo.ConstraintList()

        # **********************************************************************
        #   Declare objective function
        # **********************************************************************
        self.declare_objective(self.model)

        # **********************************************************************
        #   Built persistent model instance (if "declare_persistent" is True)
        # **********************************************************************
        if declare_persistent:
            # Store persistent solver name
            self.run_info['solver_name'] = persistent_solver

            # Call the persistent solver and assign the solver to "self.solver"
            time_persistent_start = time.time()
            self.solver = opt.SolverFactory(persistent_solver)
            # Create a "gurobipy" model object and add the model to the solver
            self.solver.set_instance(self.model)
            # Set the flag 'is_persistent_model_declared' to True
            self.is_persistent_model_declared = True
            self.log.info('    Time to set instance to persistent solver: %.2f'
                          % (time.time() - time_persistent_start))

        # **********************************************************************
        # Set flag indicating the declaration status of the model to be True
        self.is_model_declared = True

        # Store the time to build the optimization model instance
        self.run_info['model_build_time'] = int(time.time() - time_start)
        self.log.info('    Time to declare optimization model: %.2f'
                      % (time.time() - time_start))

    # ==========================================================================
    #    O P T I M I Z E
    # ==========================================================================
    def optimize(self, declare_model=True, use_clustered_data=False,
                 declare_persistent=False,
                 persistent_solver='gurobi_persistent',
                 solver='gurobi', time_limit=None, optimization_specs='',
                 results_file='results.json', tee=True):
        """
        Call the optimization routine for the EnergySystem instance.
        First, a new pyomo model instance is built (if 'declare_model'=True),
        then the model is delivered to the specified solver and finally, the
        optimization results are exported to a json-file.

        :param declare_model: Declare a new instance of the pyomo ConcreteModel
            (=> True: Call function 'declare_model') or use a previously
            declared model instance if available (=> False).
            |br| *Default: True*
        :type declare_model: bool

        :param use_clustered_data: Declare optimization model with original
            (full scale) time series (=> False) or clustered data (=> True).
            |br| *Default: False*
        :type use_clustered_data: bool

        :param declare_persistent: States if a persistent model instance should
            be formed. In this case, after model declaration a persistent solver
            instance is created and the declared model instance is assigned.
            |br| *Default: False*
        :type declare_persistent: bool

        :param persistent_solver: Name of the persistent solver to be used.
            Possible options are "gurobi_persistent" and "cplex_persistent".
            Is ignored if keyword "declare_persistent" is False.
            |br| *Default: 'gurobi_persistent'*
        :type persistent_solver: str

        :param solver: Name of the applied solver (make sure the solver is
            available on your machine and aristopy can find the path to the
            solver executables).
            |br| *Default: 'gurobi'*
        :type solver: str

        :param time_limit: Limits the total optimization time (in seconds).
            If the time limit is reached, the solver returns with the best
            currently found solution. If None is specified, solver runs until
            the default time limit is exceeded (solver specific) or another
            criteria for abortion is triggered (e.g., reached 'MIPGap').
            |br| *Default: None*
        :type time_limit: int (>0) or None

        :param optimization_specs: Additional solver parameters can be set from
            a string. To find out more about possible solver options check the
            documentation of the applied solver (e.g., see `gurobi documentation
            <https://www.gurobi.com/documentation/9.0/refman/parameters.html>`_)
            |br| E.g.: 'Threads=1 MIPGap=0.01'  |br| *Default: ''*
        :type optimization_specs: str

        :param results_file: Name of the results file (required format: .json).
            |br| *Default: 'results.json'*
        :type results_file: str or None

        :param tee: Show solver output (on screen and/or logfile).
            |br| *Default: True*
        :type tee: bool
        """
        self.log.info('Call of function "optimize"')

        solver, persistent_solver = solver.lower(), persistent_solver.lower()

        if declare_model:
            self.declare_model(declare_persistent=declare_persistent,
                               persistent_solver=persistent_solver,
                               use_clustered_data=use_clustered_data)
        elif not self.is_model_declared:
            raise ValueError('The optimization model is not declared yet. Set '
                             'the argument "declare_model" to True or call the '
                             'function "declare_model" first.')

        time_start = time.time()

        # Check input arguments
        utils.check_optimize_input(use_clustered_data,
                                   declare_persistent, persistent_solver,
                                   self.is_data_clustered, solver,
                                   time_limit, optimization_specs)

        # Add basic input arguments to the 'run_info' dict
        self.run_info['time_limit'] = time_limit
        self.run_info['optimization_specs'] = optimization_specs

        # **********************************************************************
        #   Solve model
        # **********************************************************************
        # Call a solver from the SolverFactory if the model is not persistent
        if not self.is_persistent_model_declared:
            self.run_info['solver_name'] = solver
            self.solver = opt.SolverFactory(solver)

        # Set the time limit if specified (name depends on applied solver)
        time_limit_param_names = {
            'scip': 'limits/time', 'cbc': 'seconds', 'glpk': 'tmlim',
            'gurobi': 'TimeLimit', 'gurobi_persistent': 'TimeLimit',
            'cplex': 'timelimit', 'cplex_persistent': 'timelimit',
            'baron': 'MaxTime'}
        if self.run_info['time_limit'] is not None:
            if self.is_persistent_model_declared:
                param_name = time_limit_param_names[persistent_solver]
                self.solver.options[param_name] = time_limit
            elif not self.is_persistent_model_declared and \
                    solver in time_limit_param_names.keys():
                param_name = time_limit_param_names[solver]
                self.solver.options[param_name] = time_limit

        # Set further solver options and solve the model
        self.solver.set_options(optimization_specs)
        if not self.is_persistent_model_declared:
            self.log.info('Solve non-persistent model using %s' % solver)
            solver_info = self.solver.solve(self.model, tee=tee)
        else:
            self.log.info('Solve persistent model using %s' % solver)
            solver_info = self.solver.solve(tee=tee)

        # Show the solver summary on the screen
        # solver_info.write()

        # Add selected results to the 'run_info' dict
        self.run_info['upper_bound'] = solver_info.problem.upper_bound
        self.run_info['lower_bound'] = solver_info.problem.lower_bound
        self.run_info['sense'] = str(solver_info.problem.sense)
        self.run_info['solver_status'] = str(solver_info.solver.status)
        self.run_info['termination_condition'] = str(
            solver_info.solver.termination_condition)
        self.run_info['model_solve_time'] = int(time.time() - time_start)

        self.log.info('Solve time: %d sec' % self.run_info['model_solve_time'])

        # **********************************************************************
        #   Export results to JSON
        # **********************************************************************
        # Write results to a json-file with a specified name.
        # Delete the results_file if exception is thrown.
        if results_file is not None:
            try:
                with open(results_file, 'w') as f:
                    f.write(json.dumps(self.serialize(), indent=2))
            except (ValueError or json.JSONDecodeError):
                self.log.error('Problem encountered while writing json-export.')
                if os.path.isfile(results_file):
                    os.remove(results_file)

    def export_component_configuration(self, write_to_excel=False,
                                       file_name='component_config.xlsx'):
        """
        This function collects the configuration data of all modeled components
        (results of the optimization) as pandas Series and returns them in a
        pandas DataFrame. The configuration features are (if exist):

        * the binary existence variable (utils.BI_EX),
        * the binary existence variables of modules (utils.BI_MODULE_EX)
        * the component capacity variable (utils.CAP)

        :param write_to_excel: Specify whether the component configuration
            DataFrame should be stored in an Excel-file.
            |br| *Default: False*
        :type write_to_excel: bool

        :param file_name: Name of the exported Excel-file containing the
            component configuration (if flag "write_to_excel" is True).
            |br| *Default: 'component_config.xlsx'*
        :type file_name: str

        :returns: The configuration of all components of the model instance.
        :rtype: pandas DataFrame
        """
        self.log.info('Call of function "export_component_configuration"')

        assert isinstance(write_to_excel, bool), 'input should be a boolean'
        assert isinstance(file_name, str), '"file_name" should be a string'

        for name, comp in self.components.items():
            self.component_configuration[name] = \
                comp.export_component_configuration()

        # Write configuration to Excel-file if requested:
        if write_to_excel:
            self.component_configuration.to_excel(file_name)

        return self.component_configuration

    def import_component_configuration(self, config, fix_existence=True,
                                       fix_modules=True, fix_capacity=True,
                                       store_previous_variables=True):
        """
        Function to load a pandas DataFrame with configuration specifications
        (binary existence variables and capacity variable values) and fix the
        configuration of the modeled components accordingly.

        :param config: The component configuration of the model instance
            (generated by function 'export_component_configuration').
        :type config: pandas DataFrame

        :param fix_existence: Specify whether the imported (global) binary
            component existence variables should be fixed (if available).
            |br| *Default: True*
        :type fix_existence: bool

        :param fix_modules: Specify whether the imported binary existence
            variables of the component modules should be fixed (if available).
            |br| *Default: True*
        :type fix_modules: bool

        :param fix_capacity: Specify whether the imported component capacity
            variable should be fixed or not.
            |br| *Default: True*
        :type fix_capacity: bool

        :param store_previous_variables: State whether the representation of the
            variables before applying the configuration import should be stored
            in DataFrame "variables_copy" of each component. This representation
            can be used by function "reset_component_variables" to undo changes.
            |br| *Default: True*
        :type store_previous_variables: bool
        """
        self.log.info('Call of function "import_component_configuration"')

        if not isinstance(config, pd.DataFrame):
            raise TypeError('The data needs to be imported as a pd.DataFrame!')
        utils.check_and_set_bool(fix_existence, 'fix_existence'),
        utils.check_and_set_bool(fix_modules, 'fix_modules')
        utils.check_and_set_bool(fix_capacity, 'fix_capacity')
        utils.check_and_set_bool(store_previous_variables,
                                 'store_previous_variables')

        for comp_name in config:
            if comp_name not in self.components.keys():
                self.log.warn('The configuration data of component "%s" is '
                              'imported but the component is not found in the '
                              'EnergySystem instance.' % comp_name)
            else:
                data = config[comp_name]
                comp = self.components[comp_name]
                comp.import_component_configuration(
                    data, fix_existence, fix_modules, fix_capacity,
                    store_previous_variables)

    def add_design_integer_cut_constraint(self, which_instances='all',
                                          include_existence=True,
                                          include_modules=True):
        """
        Function to add an integer cut constraint to the Concrete Model of the
        EnergySystem instance. Hereby, the currently calculated design
        solution is excluded from the solution space. Hence, in a next
        optimization run of the same model instance a different design solution
        has to be determined. The integer cuts are destroyed once a new
        optimization model is declared.

        :param which_instances: State which components should be considered
            while formulating the integer cut constraint. The argument can
            either take the string 'all' or a list of component (group) names.
            |br| *Default: 'all'*
        :type which_instances: str 'all' or list of component (group) names

        :param include_existence: State if the binary existence variables of the
            components should be considered (if available) for formulating the
            integer cut constraint.
            |br| *Default: True*
        :type include_existence: bool

        :param include_modules: State if the binary modules existence variables
            of the components should be considered (if available) for
            formulating the integer cut constraint.
            |br| *Default: True*
        :type include_modules: bool
        """
        self.log.info('Call of function "add_design_integer_cut_constraint"')

        # Check user input
        if which_instances != 'all' and not isinstance(which_instances, list):
            raise TypeError('"which_instances" takes string "all" or holds '
                            'a list of component names!')
        utils.check_and_set_bool(include_existence, 'include_existence')
        utils.check_and_set_bool(include_modules, 'include_modules')

        # Run the 'add_integer_cut_constraint' function with the desired comps
        # https://pyomo.readthedocs.io/en/latest/working_models.html
        icc_expr = 0  # init
        for name, comp in self.components.items():
            if which_instances == 'all' or name in which_instances \
                    or comp.group_name in which_instances:

                # If existence modules binaries should be used and are available
                if include_modules and comp.capacity_per_module is not None:
                    # Check if variable is constructed
                    var_bi_mod_ex = comp.variables[utils.BI_MODULE_EX]['pyomo']
                    if var_bi_mod_ex is not None:
                        # Check in every index in variable if it has a value
                        for idx in var_bi_mod_ex:
                            # Append variable to ICC expression
                            if var_bi_mod_ex[idx].value is not None:
                                if var_bi_mod_ex[idx].value < 0.5:
                                    icc_expr += var_bi_mod_ex[idx]
                                else:
                                    icc_expr += (1 - var_bi_mod_ex[idx])

                # If existence binaries should be used and are available in comp
                if include_existence and comp.has_bi_ex:
                    # ----------------------------------------------------------
                    # *** Special case: ***
                    # If a component has a binary existence variable and binary
                    # module variables, you should not use both for ICC. E.g.
                    # case: all modules do not exist (BI_MOD_EX[i]=0) and global
                    # binary can take both values (0 and 1) but nothing changes.
                    # ==> Use first module binary as global existence indicator
                    if include_modules and comp.capacity_per_module is not None:
                        continue  # skip the rest und go to next item in loop
                    elif not include_modules and comp.capacity_per_module is \
                            not None:
                        var_bi_mod_ex = comp.variables[utils.BI_MODULE_EX][
                            'pyomo']
                        if var_bi_mod_ex is not None:
                            if var_bi_mod_ex[1].value is not None:
                                if var_bi_mod_ex[1].value < 0.5:
                                    icc_expr += var_bi_mod_ex[1]
                                else:
                                    icc_expr += (1 - var_bi_mod_ex[1])
                        continue
                    # ----------------------------------------------------------
                    # Check if variable is constructed and has a value
                    var_bi_ex = comp.variables[utils.BI_EX]['pyomo']
                    if var_bi_ex is not None and var_bi_ex.value is not None:
                        # Append variable to ICC expression
                        if var_bi_ex.value < 0.5:
                            icc_expr += var_bi_ex
                        else:
                            icc_expr += (1 - var_bi_ex)

        # Add the expression to the model instance
        if self.is_model_declared:
            if self.is_persistent_model_declared:
                # add constraint to persistent solver
                self.solver.add_constraint(
                    self.model.integer_cut_constraints.add(icc_expr >= 1))
            else:  # conventional model
                self.model.integer_cut_constraints.add(icc_expr >= 1)
        else:
            raise ValueError('Integer cut constraints can only be applied if a '
                             'model has already been constructed and the '
                             'results are available!')

    def relax_integrality(self, which_instances='all', include_existence=True,
                          include_modules=True, include_time_dependent=True,
                          store_previous_variables=True):
        """
        Function to relax the integrality of the binary variables of the
        modelled components. This means binary variables are declared to be
        'NonNegativeReals with an upper bound of 1. The relaxation can be
        performed for the binary existence variable, the module existence binary
        variables and time-dependent binary variables.

        :param which_instances: State for which components the relaxation of the
            binary variables should be done. The keyword argument can either
            take the string 'all' or a list of component (group) names.
            |br| *Default: 'all'*
        :type which_instances: str 'all' or list of component (group) names

        :param include_existence: State whether the integrality of the binary
            existence variables should be relaxed (if available).
            |br| *Default: True*
        :type include_existence: bool

        :param include_modules: State whether the integrality of the binary
            modules existence variables should be relaxed (if available).
            |br| *Default: True*
        :type include_modules: bool

        :param include_time_dependent: State whether the integrality of the
            time-dependent binary variables should be relaxed. (if available).
            |br| *Default: True*
        :type include_time_dependent: bool

        :param store_previous_variables: State whether the representation of the
            variables before applying the relaxation should be stored in the
            DataFrame "variables_copy" of each component. This representation
            can be used by function "reset_component_variables" to undo changes.
            |br| *Default: True*
        :type store_previous_variables: bool
        """
        self.log.info('Call of function "relax_integrality"')

        # Check user input
        if which_instances != 'all' and not isinstance(which_instances, list):
            raise TypeError('"which_instances" takes string "all" or holds '
                            'a list of component names!')
        utils.check_and_set_bool(include_existence, 'include_existence')
        utils.check_and_set_bool(include_modules, 'include_modules')
        utils.check_and_set_bool(include_time_dependent,
                                 'include_time_dependent')
        utils.check_and_set_bool(store_previous_variables,
                                 'store_previous_variables')

        # Run the 'relax_integrality' function in the desired components
        for name, comp in self.components.items():
            if which_instances == 'all' or name in which_instances \
                    or comp.group_name in which_instances:
                comp.relax_integrality(include_existence, include_modules,
                                       include_time_dependent,
                                       store_previous_variables)

    def edit_component_variables(self, name, which_instances='all',
                                 store_previous_variables=True, **kwargs):
        """
        Method for manipulating the specifications of already defined component
        variables (e.g., change variable domain, add variable bounds, etc.).

        :param name: Name / identifier of the edited variable.
        :type name: str

        :param which_instances: State for which components the relaxation of the
            binary variables should be done. The keyword argument can either
            take the string 'all' or a list of component (group) names.
            |br| *Default: 'all'*
        :type which_instances: str 'all' or list of component (group) names

        :param store_previous_variables: State whether the representation of the
            variables before applying the relaxation should be stored in the
            DataFrame "variables_copy" of each component. This representation
            can be used by function "reset_component_variables" to undo changes.
            |br| *Default: True*
        :type store_previous_variables: bool

        :param kwargs: Additional keyword arguments for editing. Options are:
            'ub' and 'lb' to add an upper or lower variable bound, 'domain' to
            set the variable domain, and 'has_time_set' to define if the
            variable should inherit the global time set of the EnergySystem.
        """
        self.log.info('Call of function "edit_component_variables"')

        # Run the function in the desired components
        for comp in self.components.values():
            if which_instances == 'all' or comp.name in which_instances \
                    or comp.group_name in which_instances:
                comp.edit_variable(name, store_previous_variables, **kwargs)

    def reset_component_variables(self, which_instances='all'):
        """
        Function to reset the variables of the modeled components to their
        state that is stored in the component DataFrame "variables_copy".
        This includes the resetting of the DataFrame "variables" and the pyomo
        variables itself (if constructed).

        :param which_instances: State for which components the variable
            resetting should be done. The keyword argument can either take the
            string 'all' or a list of component (group) names.
            |br| *Default: 'all'*
        :type which_instances: str 'all' or list of component (group) names
        """
        self.log.info('Call of function "reset_component_variables"')
        # Check user input
        if which_instances != 'all' and not isinstance(which_instances, list):
            raise TypeError('The "which_instances" keyword can either take the '
                            'string "all" or a list of component names!')

        # Run the 'reset_variables' function in the desired components
        for name, comp in self.components.items():
            if which_instances == 'all' or name in which_instances \
                    or comp.group_name in which_instances:
                comp.reset_variables()

    # ==========================================================================
    #   Serialize the content of the class instance and its components
    # ==========================================================================
    def serialize(self):
        """
        This method collects all relevant input data and optimization results
        from the EnergySystem instance, the added components and the pyomo
        model instance. The data is arranged in an ordered dictionary, that is
        later exported as a file in JSON-format.

        :return: OrderedDict
        """

        def get_arc_variable_values():
            arc_data = {}
            for arc_name, connection in self.component_connections.items():
                # Get the pyomo block for the expanded arc
                arc_block = getattr(self.model, arc_name + '_expanded')
                # Get connected component class instances and the commodity name
                src, dest, commod = connection[0], connection[1], connection[2]
                # We can only get the commodity variable values from the arc
                # block directly, if the variables are split (not for direct
                # connection of single outlet with single inlet).
                if hasattr(arc_block, commod):
                    arc_var = str(getattr(arc_block, commod).get_values())
                else:
                    # Else: Block only holds one equality constraint (IN == OUT)
                    # --> Directly use the values of the outlet port variable of
                    # the source or the inlet port variable of the destination.
                    source_var = src.outlet_commod_and_var_names[commod]
                    arc_var = str(getattr(src.block, source_var).get_values())
                arc_data[arc_name] = arc_var
            return arc_data

        def get_added_obj_values():
            obj_values = {}
            for key, val in self.added_obj_contributions_results.items():
                obj_values[key] = pyo.value(val, exception=False)
            return obj_values

        def get_added_variable_values():
            var_dict = {}
            for var in self.added_variables.loc['pyomo']:
                var_dict[var.local_name] = str(var.get_values())
            return var_dict

        components = {}
        for name, comp in self.components.items():
            components[name] = comp.serialize()

        return OrderedDict([
            ('model_class', self.__class__.__name__),
            ('number_of_time_steps', self.number_of_time_steps),
            ('hours_per_time_step', self.hours_per_time_step),
            ('is_data_clustered', self.is_data_clustered),
            ('number_of_time_steps_per_period',
             self.number_of_time_steps_per_period),
            ('number_of_typical_periods', len(self.typical_periods)),
            ('total_number_of_periods', len(self.periods)),
            ('periods_order', str(list(self.periods_order))),
            ('period_occurrences', str(self.period_occurrences)),
            ('present value factor', self.pvf),
            ('components', components),
            ('arc_variables', get_arc_variable_values()),
            ('added_variables', get_added_variable_values()),
            ('added_objective_function_contributions', get_added_obj_values()),
            ('run_info', self.run_info)
        ])


if __name__ == '__main__':
    es = EnergySystem()



import pytest
import aristopy as ar

log = ar.Logger(default_log_level='WARNING').get_logger(__name__)


def test_storage_example():
    # Set a solver
    if hasattr(pytest, 'AVAILABLE_SOLVERS') and \
            len(pytest.AVAILABLE_SOLVERS) > 0:
        solver = pytest.AVAILABLE_SOLVERS[0]  # use first one from the list
    else:
        # if file is run as standalone test, pytest will not have the attribute
        # AVAILABLE_SOLVERS (it's set in test_000_...) --> try to use solver CBC
        solver = 'gurobi'
    log.info('The following solver is used in this test run: %s' % solver)

    es = ar.EnergySystem(number_of_time_steps=8, hours_per_time_step=1,
                         interest_rate=0, economic_lifetime=1)

    gas_src = ar.Source(ensys=es, name='gas_src', outlet=ar.Flow('F'),
                        commodity_cost=20)

    chp = ar.Conversion(ensys=es, name='chp', basic_variable='Q',
                        inlet=ar.Flow('F', 'gas_src'),
                        outlet=[ar.Flow('Q', 'bus'), ar.Flow('P', 'elec_snk')],
                        capex_per_capacity=600e3, has_operation_binary_var=True,
                        start_up_cost=10e3, min_load_rel=0.5,
                        capacity=6, user_expressions=['Q == 0.5 * F',
                                                      'P == 0.4 * F'])

    storage = ar.Storage(ensys=es, name='storage',
                         basic_variable='outlet_variable',
                         inlet=ar.Flow('Q', 'bus', 'Q_CHARGE'),
                         outlet=ar.Flow('Q', 'bus', 'Q_DISCHARGE'),
                         capex_per_capacity=1000, capacity_max=50,
                         opex_operation=1e-9, soc_initial=0.5,
                         self_discharge=0.0)

    bus = ar.Bus(es, 'bus', inlet=ar.Flow('Q', var_name='Q_IN'),
                 outlet=ar.Flow('Q', var_name='Q_OUT'), losses=0)

    elec_snk = ar.Sink(ensys=es, name='elec_snk', inlet=ar.Flow('P'),
                       commodity_revenues=42)

    heat_snk = ar.Sink(ensys=es, name='heat_snk', inlet=ar.Flow('Q', 'bus'),
                       commodity_rate_fix=ar.Series('demand',
                                                    [4, 4, 6, 6, 10, 10, 4, 4]))

    # Run the model as it is and assert the obj. fct. value
    es.optimize(declare_model=True, solver=solver, tee=False, results_file=None)

    assert es.model.Obj() == pytest.approx(-3.94438400000876e+06)

    # Perform time series aggregation and run the model again.
    # The overall result should be the same.
    es.cluster(number_of_typical_periods=3, number_of_time_steps_per_period=2)
    es.optimize(declare_model=True, use_clustered_data=True,
                solver=solver, tee=False, results_file=None)

    assert es.model.Obj() == pytest.approx(-3.94438400000876e+06)

    # Problem should be infeasible for clustered data, without using the
    # inter-period-formulation.

    # Create a clean new model!
    # Changing parameters e.g., 'storage.use_inter_period_formulation = False'
    # might work sometimes, but aristopy is not designed for that right now!

    es = ar.EnergySystem(number_of_time_steps=8, hours_per_time_step=1,
                         interest_rate=0, economic_lifetime=1)
    gas_src = ar.Source(ensys=es, name='gas_src', outlet=ar.Flow('F'),
                        commodity_cost=20)
    chp = ar.Conversion(ensys=es, name='chp', basic_variable='Q',
                        inlet=ar.Flow('F', 'gas_src'),
                        outlet=[ar.Flow('Q', 'bus'), ar.Flow('P', 'elec_snk')],
                        capex_per_capacity=600e3, has_operation_binary_var=True,
                        start_up_cost=10e3, min_load_rel=0.5,
                        capacity=6, user_expressions=['Q == 0.5 * F',
                                                      'P == 0.4 * F'])
    storage = ar.Storage(ensys=es, name='storage',
                         basic_variable='outlet_variable',
                         inlet=ar.Flow('Q', 'bus', 'Q_CHARGE'),
                         outlet=ar.Flow('Q', 'bus', 'Q_DISCHARGE'),
                         capex_per_capacity=1000, capacity_max=50,
                         opex_operation=1e-9, soc_initial=0.5,
                         self_discharge=0.0,
                         use_inter_period_formulation=False)  # FALSE!

    bus = ar.Bus(es, 'bus', inlet=ar.Flow('Q', var_name='Q_IN'),
                 outlet=ar.Flow('Q', var_name='Q_OUT'), losses=0)
    elec_snk = ar.Sink(ensys=es, name='elec_snk', inlet=ar.Flow('P'),
                       commodity_revenues=42)
    heat_snk = ar.Sink(ensys=es, name='heat_snk', inlet=ar.Flow('Q', 'bus'),
                       commodity_rate_fix=ar.Series('demand',
                                                    [4, 4, 6, 6, 10, 10, 4, 4]))

    es.cluster(number_of_typical_periods=3, number_of_time_steps_per_period=2)
    es.optimize(declare_model=True, use_clustered_data=True,
                solver=solver, tee=False, results_file=None)

    assert es.run_info['termination_condition'] in [
        'infeasible', 'unbounded', 'infeasibleOrUnbounded']


if __name__ == '__main__':
    test_storage_example()
